import unittest
from typing import List, Union

import pydantic
import rdflib
from pydantic import HttpUrl, ValidationError

from ontolutils import Thing, urirefs, namespaces
from ontolutils import set_logging_level
from ontolutils import typing
from ontolutils.typing import ResourceType

set_logging_level('WARNING')


class TestTypeing(unittest.TestCase):

    def test_blank_node(self):
        class MyModel(pydantic.BaseModel):
            blank_node: typing.BlankNodeType

        with self.assertRaises(pydantic.ValidationError):
            MyModel(blank_node='b1')

        with self.assertRaises(pydantic.ValidationError):
            MyModel(blank_node='_b1')

        MyModel(blank_node='_:b1')

    def test_resource_type(self):
        @namespaces(dcterms="http://purl.org/dc/terms/",
                    dcat="http://www.w3.org/ns/dcat#")
        @urirefs(Dataset='dcat:Dataset',
                 hasVersion='dcterms:hasVersion')
        class Dataset(Thing):
            """
            A dataset class.
            """
            hasVersion: Union[ResourceType, List[ResourceType]]

        with self.assertRaises(ValidationError):
            Dataset(
                label="Test ds@en",
                hasVersion="v1.0"
            )

        ds = Dataset(
            label="Test ds@en",
            hasVersion="http://example.org/resource/version/v1.0"
        )
        ttl = ds.serialize("ttl")
        self.assertEqual(ttl, """@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a dcat:Dataset ;
    rdfs:label "Test ds"@en ;
    dcterms:hasVersion <http://example.org/resource/version/v1.0> .

""")

        ds = Dataset(
            label="Test ds@en",
            hasVersion=["http://example.org/resource/version/v1.0",
                        "http://example.org/resource/version/v2.0"]
        )
        ttl = ds.serialize("ttl")
        self.assertEqual(ttl, """@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a dcat:Dataset ;
    rdfs:label "Test ds"@en ;
    dcterms:hasVersion <http://example.org/resource/version/v1.0>,
        <http://example.org/resource/version/v2.0> .

""")

        version = Thing(
            id="http://example.org/resource/version/v1.0",
            label="v1.0"
        )

        ds = Dataset(
            label="Test ds@en",
            hasVersion=version
        )
        ttl = ds.serialize("ttl")
        self.assertEqual(ttl, """@prefix dcat: <http://www.w3.org/ns/dcat#> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

<http://example.org/resource/version/v1.0> a owl:Thing ;
    rdfs:label "v1.0" .

[] a dcat:Dataset ;
    rdfs:label "Test ds"@en ;
    dcterms:hasVersion <http://example.org/resource/version/v1.0> .

""")

        ds = Dataset(
            label="Test ds@en",
            hasVersion=rdflib.URIRef("http://example.org/resource/version/v1.0")
        )
        self.assertEqual("http://example.org/resource/version/v1.0", ds.hasVersion)
        ds = Dataset(
            label="Test ds@en",
            hasVersion=HttpUrl("http://example.org/resource/version/v1.0")
        )
        self.assertEqual("http://example.org/resource/version/v1.0", ds.hasVersion)



import datetime
import json
import logging
import unittest
from itertools import count
from typing import Optional, List, Union

import pydantic
import rdflib
import yaml
from pydantic import EmailStr, model_validator, HttpUrl
from pydantic import ValidationError
from pydantic import field_validator, Field
from rdflib.plugins.shared.jsonld.context import Context
from typing_extensions import Annotated

import ontolutils
from ontolutils import SCHEMA
from ontolutils import Thing, urirefs, namespaces, build, Property
from ontolutils import as_id
from ontolutils import get_urirefs, get_namespaces, set_config
from ontolutils import set_logging_level
from ontolutils.classes import decorator
from ontolutils.classes.thing import resolve_iri, LangString
from ontolutils.classes.utils import split_URIRef

LOG_LEVEL = logging.DEBUG


class TestNamespaces(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        logger = logging.getLogger('ontolutils')
        self.INITIAL_LOG_LEVEL = logger.level

        set_logging_level(LOG_LEVEL)

        assert logger.level == LOG_LEVEL

    def tearDown(self):
        set_logging_level(self.INITIAL_LOG_LEVEL)
        assert logging.getLogger('ontolutils').level == self.INITIAL_LOG_LEVEL

    def test_model_fields(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 name='foaf:lastName',
                 age='foaf:age')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            name: str = Field(default=None, alias="lastName")  # name is synonymous to lastName
            age: int = None
            special_field: Optional[str] = None

            @field_validator('special_field', mode="before")
            @classmethod
            def _special_field(cls, value):
                assert value == "special_string", f"Special field must be 'special_string' not {value}"
                return value

        agent = Agent(name='John Doe', age=23)
        self.assertEqual(agent.name, 'John Doe')
        self.assertEqual(agent.age, 23)

        # extra fields are allowed and either the model field or its uriref can be used
        agent = Agent(lastName='Doe', age=23)
        self.assertEqual(agent.name, 'Doe')
        self.assertEqual(agent.lastName, 'Doe')
        self.assertEqual(agent.age, 23)

        agent = Agent(age=23)
        self.assertEqual(agent.name, None)
        self.assertEqual(agent.age, 23)

        # property assignment should fail:
        with self.assertRaises(pydantic.ValidationError):
            agent.age = "invalid"

        with self.assertRaises(pydantic.ValidationError):
            agent.special_field = "invalid"
        self.assertEqual(agent.special_field, None)

        agent.special_field = "special_string"
        self.assertEqual(agent.special_field, "special_string")

    def test_resolve_iri(self):
        ret = resolve_iri('foaf:age', context=Context(source={'foaf': 'http://xmlns.com/foaf/0.1/'}))
        self.assertEqual(ret, 'http://xmlns.com/foaf/0.1/age')

        ret = resolve_iri('age', context=Context(source={'foaf': 'http://xmlns.com/foaf/0.1/'}))
        self.assertEqual(ret, None)

        ret = resolve_iri('age', context=Context(source={'age': 'http://xmlns.com/foaf/0.1/age'}))
        self.assertEqual(ret, 'http://xmlns.com/foaf/0.1/age')

        ret = resolve_iri('label', context=Context(source={'age': 'http://xmlns.com/foaf/0.1/age'}))
        self.assertEqual(ret, 'http://www.w3.org/2000/01/rdf-schema#label')

        ret = resolve_iri('label',
                          context=Context(source={'label': {'@id': 'http://www.w3.org/2000/01/rdf-schema#label'}}))
        self.assertEqual(ret, 'http://www.w3.org/2000/01/rdf-schema#label')

        ret = resolve_iri('prefix:label', Context(source={}))
        self.assertEqual(ret, None)

    def test_split_URIRef(self):
        self.assertListEqual(split_URIRef(rdflib.URIRef('https://example.com/')),
                             ['https://example.com/', ''])
        self.assertListEqual(split_URIRef(rdflib.URIRef('https://example.com/#test')),
                             ['https://example.com/#', 'test'])
        self.assertListEqual(split_URIRef(rdflib.URIRef('https://example.com/test')),
                             ['https://example.com/', 'test'])
        self.assertListEqual(split_URIRef(rdflib.URIRef('https://example.com/test#test')),
                             ['https://example.com/test#', 'test'])
        self.assertListEqual(split_URIRef(rdflib.URIRef('https://example.com/test:123')),
                             ['https://example.com/', 'test:123'])

    def test_thing_custom_prop(self):
        """It is helpful to have the properties equal to the urirefs keys,
        however, this should not be required!"""

        @namespaces(foaf='http://xmlns.com/foaf/0.1/',
                    schema='https://www.schema.org/')
        @urirefs(Affiliation='prov:Affiliation',
                 name='schema:name')
        class Affiliation(Thing):
            name: str

        @namespaces(foaf='http://xmlns.com/foaf/0.1/',
                    prov='https://www.w3.org/ns/prov#')
        @urirefs(Person='prov:Person',
                 first_name='foaf:firstName',
                 lastName='foaf:lastName',
                 age='foaf:age')
        class Person(Thing):
            first_name: str = Field(default=None, alias='firstName')
            lastName: str
            age: int = None
            affiliation: Affiliation = None

        p = Person(first_name='John', lastName='Doe', age=23)
        person_json = p.model_dump_jsonld(resolve_keys=False)
        self.assertEqual(json.loads(person_json)['first_name'], 'John')
        person_json = p.model_dump_jsonld(resolve_keys=True)
        self.assertEqual(json.loads(person_json)['foaf:firstName'], 'John')

        p_from_jsonld = Person.from_jsonld(data=p.model_dump_jsonld(resolve_keys=True), limit=1)
        self.assertEqual(p_from_jsonld.first_name, 'John')
        self.assertEqual(p_from_jsonld.lastName, 'Doe')
        self.assertEqual(p_from_jsonld.age, 23)

        p_from_jsonld = Person.from_jsonld(data=p.model_dump_jsonld(resolve_keys=False), limit=1)
        self.assertEqual(p_from_jsonld.first_name, 'John')
        self.assertEqual(p_from_jsonld.lastName, 'Doe')
        self.assertEqual(p_from_jsonld.age, 23)

        p_from_jsonld = Person.from_jsonld(data=p.model_dump_jsonld(resolve_keys=True), limit=None)
        self.assertEqual(p_from_jsonld[0].first_name, 'John')
        self.assertEqual(p_from_jsonld[0].lastName, 'Doe')
        self.assertEqual(p_from_jsonld[0].age, 23)

        p_from_jsonld = Person.from_jsonld(data=p.model_dump_jsonld(resolve_keys=False), limit=None)
        self.assertEqual(p_from_jsonld[0].first_name, 'John')
        self.assertEqual(p_from_jsonld[0].lastName, 'Doe')
        self.assertEqual(p_from_jsonld[0].age, 23)

        # add additional non-urirefs property:
        p.height = 183
        self.assertEqual(p.height, 183)
        self.assertEqual(json.loads(p.model_dump_jsonld(resolve_keys=True))['height'], 183)
        p183_from_jsonld = Person.from_jsonld(data=p.model_dump_jsonld(resolve_keys=True), limit=1)
        with self.assertRaises(AttributeError):  # height is not defined!
            p183_from_jsonld.height

    def test_from_jsonld_for_nested_objects(self):
        @namespaces(prov='https://www.w3.org/ns/prov#')
        @urirefs(A='prov:A',
                 name='prov:name')
        class A(Thing):
            name: str = None

        @namespaces(prov='https://www.w3.org/ns/prov#')
        @urirefs(B='prov:B',
                 a='prov:a')
        class B(Thing):
            a: A = None

        @namespaces(prov='https://www.w3.org/ns/prov#')
        @urirefs(C='prov:C',
                 b='prov:b')
        class C(Thing):
            b: B = None

        @namespaces(prov='https://www.w3.org/ns/prov#')
        @urirefs(D='prov:D',
                 c='prov:c')
        class D(Thing):
            c: C = None

        aj = A(name="myname").model_dump_jsonld()
        an = A.from_jsonld(data=aj, limit=1)
        self.assertEqual("myname", an.name)

        bj = B(a=(A(name="myname"))).model_dump_jsonld()
        bn = B.from_jsonld(data=bj, limit=1)
        self.assertEqual("myname", bn.a.name)

        cj = C(b=B(a=(A(name="myname")))).model_dump_jsonld()
        cn = C.from_jsonld(data=cj, limit=1)
        self.assertEqual("myname", cn.b.a.name)

        dj = D(c=C(b=B(a=(A(name="myname"))))).model_dump_jsonld()
        dn = D.from_jsonld(data=dj, limit=1)
        self.assertEqual("myname", dn.c.b.a.name)

    def test_sort_classes(self):
        thing1 = Thing(label='Thing 1', id='_:1')
        thing2 = Thing(label='Thing 2', id='_:2')
        self.assertFalse(thing1 > thing2)
        with self.assertRaises(TypeError):
            thing1 < 4
        thing1 = Thing(label='Thing 1', id='https://example.com/thing1')
        thing2 = Thing(label='Thing 2', id='https://example.com/thing2')
        self.assertTrue(thing1 < thing2)

    def test_language_string0(self):
        thing_en = Thing(label=rdflib.Literal(lexical_or_value='a thing.', lang='en'))
        ttl = thing_en.model_dump_ttl()
        self.assertEqual(ttl, """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a owl:Thing ;
    rdfs:label "a thing."@en .

""")

    def test_language_string1(self):
        self.assertNotEqual("x", LangString(value='upward', lang=None))
        self.assertEqual("a thing", LangString(value="a thing", lang="en").value)
        self.assertEqual("a thing", LangString(value="a thing").value)
        self.assertEqual("a thing", LangString(value="a thing"))
        self.assertEqual("a thing", LangString(value="a thing", lang="en"))
        self.assertNotEqual("a thing", LangString(value="another thing").value)
        self.assertNotEqual("a thing", LangString(value="another thing", lang="en").value)

        self.assertNotEqual(LangString(value="Hello", lang="en"), "Hello@fr")
        thing_en = Thing(label=LangString(value='a thing.', lang='en'))
        with set_config(show_lang_in_str=True):
            ttl = thing_en.model_dump_ttl()
            self.assertEqual(ttl, """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a owl:Thing ;
    rdfs:label "a thing."@en .

""")
        with set_config(show_lang_in_str=False):
            self.assertEqual(str(LangString(value="a thing", lang="en")), "a thing")
        with set_config(show_lang_in_str=True):
            self.assertEqual(str(LangString(value="a thing", lang="en")), "a thing@en")
        with set_config(show_lang_in_str=True):
            self.assertEqual(str(LangString(value="a thing")), "a thing")

    def test_language_string2(self):
        thing_en = Thing(
            label=[LangString(value='a thing.', lang='en'),
                   LangString(value='ein Ding.', lang='de'), ]
        )
        ttl = thing_en.model_dump_ttl()
        self.assertEqual(ttl, """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a owl:Thing ;
    rdfs:label "ein Ding."@de,
        "a thing."@en .

""")

    def test_language_string3(self):
        thing_en = Thing(label="deutsch@de")
        ttl = thing_en.model_dump_ttl()
        self.assertEqual(ttl, """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a owl:Thing ;
    rdfs:label "deutsch"@de .

""")

        #         thing_en = Thing(label=LangString(value='2025-01-01', datatype=XSD.date))
        #         ttl = thing_en.model_dump_ttl()
        #         self.assertEqual(ttl, """@prefix owl: <http://www.w3.org/2002/07/owl#> .
        # @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
        # @prefix xsd: <http://www.w3.org/2001/XMLSchema#> .
        #
        # [] a owl:Thing ;
        #     rdfs:label "2025-01-01"^^xsd:date .
        #
        # """)

    def test_language_string4(self):
        thing_en = Thing(label=[
            "a thing.@en",
            "ein Ding@de"
        ])
        ttl = thing_en.model_dump_ttl()
        self.assertEqual(ttl, """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a owl:Thing ;
    rdfs:label "ein Ding"@de,
        "a thing."@en .

""")

    def test_lang_string_yaml_support(self):
        ls = LangString(value="hallo", lang="de")
        with open("output.yaml", "w", encoding="utf-8") as f:
            yaml.dump(ls, f, allow_unicode=True)
        with open("output.yaml", "r", encoding="utf-8") as f:
            ls2 = yaml.load(f, Loader=yaml.Loader)
        self.assertEqual(ls, ls2)

    def test__repr_html_(self):
        thing = Thing(label='Thing 1')
        self.assertEqual(thing._repr_html_(), f'Thing(id={thing.id}, label=Thing 1)')

    def test_serialize_date(self):
        @namespaces(dcterms="http://purl.org/dc/terms/")
        @urirefs(MyThing='foaf:MyThing',
                 created='dcterms:created')
        class MyThing(Thing):
            created: datetime.datetime = None

        mything = MyThing(created=datetime.datetime(year=2025, month=10, day=1, hour=12, minute=30))
        ttl = mything.serialize("ttl")
        self.assertEqual(ttl, """@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a <foaf:MyThing> ;
    dcterms:created "2025-10-01T12:30:00"^^xsd:dateTime .

""")
        mything = MyThing(created=datetime.datetime(year=2025, month=10, day=1))
        ttl = mything.serialize("ttl")
        self.assertEqual(ttl, """@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a <foaf:MyThing> ;
    dcterms:created "2025-10-01"^^xsd:date .

""")

    def test_thing_get_jsonld_dict(self):
        with self.assertRaises(TypeError):
            _ = Thing(id=1, label='Thing 1')
        with self.assertRaises(pydantic.ValidationError):
            _ = Thing(id="1", label='Thing 1')

        thing = Thing(id='https://example.org/TestThing', label='Test Thing', numerical_value=1.0,
                      dt=datetime.datetime(2021, 1, 1))
        with self.assertRaises(TypeError):
            thing.get_jsonld_dict(context=1, base_uri=None)

        thing_dict = thing.get_jsonld_dict(
            resolve_keys=True,
            exclude_none=True,
            base_uri=None,
            context=None
        )
        self.assertIsInstance(thing_dict, dict)
        self.assertDictEqual(
            thing_dict['@context'],
            {'owl': 'http://www.w3.org/2002/07/owl#',
             'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
             'schema': 'https://schema.org/',
             'skos': 'http://www.w3.org/2004/02/skos/core#',
             'dcterms': 'http://purl.org/dc/terms/'
             }
        )

        self.assertEqual(thing_dict['@id'], 'https://example.org/TestThing')
        self.assertEqual(thing_dict['rdfs:label'], 'Test Thing')
        self.assertEqual(thing_dict['@type'], 'owl:Thing')

    def test_decorator(self):
        self.assertTrue(decorator._is_http_url('https://example.com/'))
        self.assertFalse(decorator._is_http_url('example.com/'))
        self.assertFalse(decorator._is_http_url('http:invalid.123'))

    def test_model_dump_jsonld(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 mbox='foaf:mbox',
                 age='foaf:age')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            mbox: EmailStr = None
            age: int = None

        agent = Agent(
            label='Agent 1',
            mbox='my@email.com',
            age=23,
        )

        with self.assertRaises(pydantic.ValidationError):
            agent.mbox = 4.5
            agent.model_validate(agent.model_dump())
        agent.mbox = 'my@email.com'
        jsonld_str1 = agent.model_dump_jsonld(rdflib_serialize=False)
        self.assertTrue('@id' in json.loads(jsonld_str1))

        self.assertIsInstance(json.loads(jsonld_str1)['foaf:age'], int)

        jsonld_str2 = agent.model_dump_jsonld(rdflib_serialize=True)  # will assign blank node! Pop it later
        jsonld_str2_dict = json.loads(jsonld_str2)
        self.assertDictEqual(
            json.loads(jsonld_str1),
            jsonld_str2_dict
        )

        agent1_dict = json.loads(jsonld_str1)
        agent1_dict.pop('@id')

        agent2_dict = jsonld_str2_dict
        agent2_dict.pop('@id')

        self.assertDictEqual(agent1_dict,
                             agent2_dict)

        # jsonld_str2_dict.pop('@id')
        # self.assertEqual(
        #     json.loads(jsonld_str1),
        #     jsonld_str2_dict
        # )

        # serialize with a "@import"
        jsonld_str3 = agent.model_dump_jsonld(
            rdflib_serialize=False,
            context={
                '@import': 'https://git.rwth-aachen.de/nfdi4ing/metadata4ing/metadata4ing/-/raw/master/m4i_context.jsonld'
            }
        )
        jsonld_str3_dict = json.loads(jsonld_str3)
        self.assertEqual(
            jsonld_str3_dict['@context']['@import'],
            'https://git.rwth-aachen.de/nfdi4ing/metadata4ing/metadata4ing/-/raw/master/m4i_context.jsonld'
        )

        def base_uri_generator():
            return f"https://example.org/agents/{rdflib.BNode()}"

        with set_config(blank_id_generator=base_uri_generator):
            new_agent = Agent(
                label='Agent new',
                mbox='new_agent@email.com',
                age=33, )
            jsonld_with_base_uri = new_agent.model_dump_jsonld()
        self.assertTrue(
            Agent.from_jsonld(data=jsonld_with_base_uri, limit=1).id.startswith("https://example.org/agents/"))

        new_agent = Agent(
            label='Agent new',
            mbox='new_agent@email.com',
            age=33, )
        jsonld_with_blank_node_base_uri = new_agent.model_dump_jsonld()
        self.assertTrue(Agent.from_jsonld(data=jsonld_with_blank_node_base_uri, limit=1).id.startswith("_:"))

        jsonld_with_base_uri = new_agent.model_dump_jsonld(base_uri="https://example.org/agents/")
        self.assertTrue(
            Agent.from_jsonld(data=jsonld_with_base_uri, limit=1).id.startswith("https://example.org/agents/"))

        new_agent = Agent(
            id="https://example.org/agents/new_agent/123",
            label='Agent new',
            mbox='new_agent@email.com',
            age=33, )
        new_agent_jsonld = new_agent.model_dump_jsonld()
        self.assertEqual(Agent.from_jsonld(data=new_agent_jsonld, limit=1).id,
                         "https://example.org/agents/new_agent/123")

        new_agent_jsonld = new_agent.model_dump_jsonld(base_uri="https://example.org/agents/")
        self.assertEqual(Agent.from_jsonld(data=new_agent_jsonld, limit=1).id,
                         "https://example.org/agents/new_agent/123")

        new_agent = Agent(
            id="_:123",
            label='Agent new',
            mbox='new_agent@email.com',
            age=33, )
        new_agent_jsonld = new_agent.model_dump_jsonld(base_uri="https://example.org/agents/")
        self.assertEqual(Agent.from_jsonld(data=new_agent_jsonld, limit=1).id,
                         "https://example.org/agents/123")

    def test_model_dump_ttl(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 mbox='foaf:mbox',
                 age='foaf:age')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            mbox: EmailStr = None
            age: int = None

        agent = Agent(
            label='Agent 1',
            mbox='my@email.com',
            age=23,
        )
        self.assertEqual(
            """@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a foaf:Agent ;
    rdfs:label "Agent 1" ;
    foaf:age 23 ;
    foaf:mbox "my@email.com" .

""",
            agent.model_dump_ttl()
        )

    def test_model_dump_jsonld_and_load_with_import(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 mbox='foaf:mbox')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            mbox: EmailStr = None

        agent = Agent(
            label='Agent 1',
            mbox='my@email.com'
        )
        self.assertNotEqual(agent.id, None)
        ns = agent.namespaces
        jsonld_string = agent.model_dump_jsonld(
            context={
                "@import": 'https://raw.githubusercontent.com/matthiasprobst/pivmeta/main/pivmeta_context.jsonld'
            }
        )
        self.assertDictEqual(agent.namespaces, ns)
        self.assertTrue('@import' in json.loads(jsonld_string)['@context'])
        loaded_agent = Agent.from_jsonld(data=jsonld_string, limit=1)
        self.assertDictEqual(loaded_agent.namespaces, ns)
        self.assertEqual(loaded_agent.mbox, agent.mbox)

        # do the same with thing:
        thing = Thing.from_jsonld(data=jsonld_string, limit=1)
        self.assertEqual(thing.label, LangString(value="Agent 1"))
        self.assertTrue(thing.id.startswith('_:'))
        _id = thing.id

    def test_schema_http(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/",
                    schema="https://schema.org/")
        @urirefs(Agent='foaf:Agent',
                 name='schema:name')
        class Agent(Thing):
            name: str

        agent = Agent(name='John Doe')
        self.assertEqual(agent.name, 'John Doe')
        agent_jsonld = agent.model_dump_jsonld()
        with self.assertWarns(UserWarning):
            agent.from_jsonld(data=agent_jsonld.replace('https://schema', 'http://schema'),
                              limit=1)

    def test_model_dump_jsonld_nested(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 mbox='foaf:mbox')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            mbox: EmailStr = None

        @namespaces(schema="https://schema.org/")
        @urirefs(Organization='prov:Organization')
        class Organization(Agent):
            """Pydantic Model for https://www.w3.org/ns/prov/Agent"""

        @namespaces(schema="https://schema.org/")
        @urirefs(Person='foaf:Person',
                 affiliation='schema:affiliation')
        class Person(Agent):
            firstName: str = None
            affiliation: Union[str, HttpUrl, Organization, List[Union[str, HttpUrl, Organization]]] = None

        person = Person(
            label='Person 1',
            affiliation=Organization(
                label='Organization 1'
            ),
        )
        jsonld_str = person.model_dump_jsonld(resolve_keys=True)
        jsonld_dict = json.loads(jsonld_str)

        self.assertEqual(jsonld_dict['schema:affiliation']['@type'], 'prov:Organization')
        self.assertEqual(jsonld_dict['schema:affiliation']['rdfs:label'], 'Organization 1')
        self.assertEqual(jsonld_dict['rdfs:label'], 'Person 1')
        self.assertEqual(jsonld_dict['@type'], 'foaf:Person')

        person = Person(
            label='Person 1',
            affiliation='https://schema:Organization/123',
        )

        print(person.serialize("ttl"))

        person = Person(
            label='Person 1',
            affiliation=[
                'https://schema.org/Organization/1',
                'https://schema.org/Organization/2'
            ],
        )

        print(person.serialize("ttl"))

    def test_prov(self):
        @namespaces(prov="https://www.w3.org/ns/prov#",
                    foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='prov:Agent',
                 mbox='foaf:mbox')
        class Agent(Thing):
            """Pydantic Model for https://www.w3.org/ns/prov#Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            mbox: EmailStr = None  # foaf:mbox

        with self.assertRaises(pydantic.ValidationError):
            agent = Agent(mbox='123')

        agent = Agent(mbox='m@email.com')
        self.assertEqual(agent.mbox, 'm@email.com')
        self.assertEqual(agent.mbox, agent.model_dump()['mbox'])
        self.assertEqual(Agent.iri(), 'https://www.w3.org/ns/prov#Agent')
        self.assertEqual(Agent.iri(compact=True), 'prov:Agent')
        self.assertEqual(Agent.iri('mbox'), 'http://xmlns.com/foaf/0.1/mbox')
        self.assertEqual(Agent.iri('mbox', compact=True), 'foaf:mbox')

    def test_use_as_id(self):
        @namespaces(prov="https://www.w3.org/ns/prov#",
                    foaf="http://xmlns.com/foaf/0.1/",
                    m4i="http://w3id.org/nfdi4ing/metadata4ing#"
                    )
        @urirefs(Person='prov:Person',
                 firstName='foaf:firstName',
                 lastName='foaf:lastName',
                 orcidId='m4i:orcidId',
                 mbox='foaf:mbox')
        class Person(Thing):
            firstName: str
            lastName: str = None
            mbox: EmailStr = None
            orcidId: str = Field(default=None, alias="orcid_id")

            @model_validator(mode="before")
            def _change_id(self):
                return as_id(self, "orcidId")

        p = Person(
            id="_:cde4c79c-21f2-4ab7-b01d-28de6e4aade4",
            firstName='John',
            lastName='Doe',
            orcidId='https://orcid.org/0000-0001-8729-0482', )
        jsonld = {
            "@context": {
                'm4i': 'http://w3id.org/nfdi4ing/metadata4ing#',
                "owl": "http://www.w3.org/2002/07/owl#",
                "prov": "https://www.w3.org/ns/prov#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                "dcterms": "http://purl.org/dc/terms/",
                'schema': 'https://schema.org/',
                "skos": "http://www.w3.org/2004/02/skos/core#",
                "foaf": "http://xmlns.com/foaf/0.1/",
            },
            "@type": "prov:Person",
            "foaf:firstName": "John",
            "foaf:lastName": "Doe",
            "m4i:orcidId": {"@id": "https://orcid.org/0000-0001-8729-0482"},
            "@id": "_:cde4c79c-21f2-4ab7-b01d-28de6e4aade4",
        }

        self.assertDictEqual(json.loads(p.model_dump_jsonld()),
                             jsonld)

        p_without_http_orcid = Person(
            firstName='John',
            lastName='Doe',
            orcidId='0000-0001-8729-0482')
        self.assertEqual(json.loads(p_without_http_orcid.model_dump_jsonld())["@id"],
                         p_without_http_orcid.id)
        self.assertEqual(json.loads(p_without_http_orcid.model_dump_jsonld())["m4i:orcidId"],
                         '0000-0001-8729-0482')

        p = Person(
            firstName='John',
            lastName='Doe',
            orcidId='https://orcid.org/0000-0001-8729-0482', )
        jsonld = {
            "@context": {
                'm4i': 'http://w3id.org/nfdi4ing/metadata4ing#',
                "owl": "http://www.w3.org/2002/07/owl#",
                "prov": "https://www.w3.org/ns/prov#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                'schema': 'https://schema.org/',
                "dcterms": "http://purl.org/dc/terms/",
                "skos": "http://www.w3.org/2004/02/skos/core#",
                "foaf": "http://xmlns.com/foaf/0.1/",
            },
            "@type": "prov:Person",
            "foaf:firstName": "John",
            "foaf:lastName": "Doe",
            "m4i:orcidId": {"@id": "https://orcid.org/0000-0001-8729-0482"},
            "@id": "https://orcid.org/0000-0001-8729-0482",
        }

        self.assertDictEqual(json.loads(p.model_dump_jsonld()),
                             jsonld)

    def test_use_as_id_V2(self):
        @namespaces(schema="https://schema.org/",
                    foaf="http://xmlns.com/foaf/0.1/",
                    )
        @urirefs(Orga='prov:Organization',
                 name="schema:name",
                 identifier="schema:identifier",
                 mbox='foaf:mbox')
        class Orga(Thing):
            identifier: str = Field(default=None, alias="identifier")
            name: str = Field(default=None, alias="name")
            mbox: EmailStr = None

            @model_validator(mode="before")
            def _change_id(self):
                return as_id(self, "identifier")

        @namespaces(prov="https://www.w3.org/ns/prov#",
                    foaf="http://xmlns.com/foaf/0.1/",
                    m4i="http://w3id.org/nfdi4ing/metadata4ing#"
                    )
        @urirefs(Person='prov:Person',
                 firstName='foaf:firstName',
                 lastName='foaf:lastName',
                 orcidId='m4i:orcidId',
                 mbox='foaf:mbox')
        class Person(Thing):
            firstName: str
            lastName: str = None
            mbox: EmailStr = None
            orcidId: str = Field(default=None, alias="orcid_id")
            affiliation: Orga = None

            @model_validator(mode="before")
            def _change_id(self):
                return as_id(self, "orcidId")

        p = Person(
            id="_:cde4c79c-21f2-4ab7-b01d-28de6e4aade4",
            firstName='John',
            lastName='Doe',
            orcidId='https://orcid.org/0000-0001-8729-0482',
            affiliation=Orga(identifier='https://example.org/123', name='Orga 1')
        )
        # Person was created with an explicit ID
        self.assertEqual(p.id, "_:cde4c79c-21f2-4ab7-b01d-28de6e4aade4")

        p = Person(
            firstName='John',
            lastName='Doe',
            orcidId='https://orcid.org/0000-0001-8729-0482',
            affiliation=Orga(identifier='https://example.org/123', name='Orga 1')
        )

        pdict = json.loads(p.model_dump_jsonld())
        self.assertEqual(pdict['@id'], 'https://orcid.org/0000-0001-8729-0482')
        self.assertEqual(pdict['affiliation']["@id"], 'https://example.org/123')

    def test_use_as_id_V3(self):
        @namespaces(schema="https://schema.org/",
                    foaf="http://xmlns.com/foaf/0.1/",
                    )
        @urirefs(Orga='prov:Organization',
                 name="schema:name",
                 identifier="schema:identifier",
                 mbox='foaf:mbox')
        class Orga(Thing):
            identifier: str = Field(default=None, alias="identifier")
            name: str = Field(default=None, alias="name")
            mbox: EmailStr = None

            @model_validator(mode="after")
            def _change_id(self):
                return as_id(self, "identifier")

        @namespaces(prov="https://www.w3.org/ns/prov#",
                    foaf="http://xmlns.com/foaf/0.1/",
                    m4i="http://w3id.org/nfdi4ing/metadata4ing#"
                    )
        @urirefs(Person='prov:Person',
                 firstName='foaf:firstName',
                 lastName='foaf:lastName',
                 orcidId='m4i:orcidId',
                 mbox='foaf:mbox')
        class Person(Thing):
            firstName: str
            lastName: str = None
            mbox: EmailStr = None
            orcidId: str = Field(default=None, alias="orcid_id")
            affiliation: Orga = None

            @model_validator(mode="before")
            def _change_id(self):
                return as_id(self, "orcidId")

        with self.assertRaises(ValueError):
            Person(
                id="_:cde4c79c-21f2-4ab7-b01d-28de6e4aade4",
                firstName='John',
                lastName='Doe',
                orcidId='https://orcid.org/0000-0001-8729-0482',
                affiliation=Orga(identifier='123', name='Orga 1')
            )

    def test_update_namespace_and_uri(self):
        class CustomPerson(Thing):
            pass

        mt = CustomPerson()
        # custom person has no
        self.assertDictEqual(mt.urirefs, get_urirefs(Thing))
        self.assertDictEqual(mt.urirefs,
                             {'Thing': 'owl:Thing', 'closeMatch': 'skos:closeMatch', 'exactMatch': 'skos:exactMatch',
                              'label': 'rdfs:label', 'about': 'schema:about', 'relation': 'dcterms:relation'})
        self.assertDictEqual(mt.namespaces, get_namespaces(Thing))
        self.assertDictEqual(mt.namespaces, {'owl': 'http://www.w3.org/2002/07/owl#',
                                             'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
                                             'schema': 'https://schema.org/',
                                             'skos': 'http://www.w3.org/2004/02/skos/core#',
                                             'dcterms': 'http://purl.org/dc/terms/'})

        mt = CustomPerson(first_name='John', last_name='Doe')
        with self.assertRaises(AttributeError):
            mt.namespaces = 'http://xmlns.com/foaf/0.1/'
        with self.assertRaises(AttributeError):
            mt.urirefs = 'foaf:lastName'

        mt.namespaces['foaf'] = 'http://xmlns.com/foaf/0.1/'
        mt.urirefs['first_name'] = 'foaf:firstName'
        mt.urirefs['last_name'] = 'foaf:lastName'

        ref_jsonld = {
            "@context": {
                "owl": "http://www.w3.org/2002/07/owl#",
                "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
                'schema': 'https://schema.org/',
                "dcterms": "http://purl.org/dc/terms/",
                "skos": "http://www.w3.org/2004/02/skos/core#",
                "foaf": "http://xmlns.com/foaf/0.1/"
            },
            "@type": "CustomPerson",
            "foaf:firstName": "John",
            "foaf:lastName": "Doe"
        }
        jsonld_dict = json.loads(mt.model_dump_jsonld())
        jsonld_dict.pop('@id', None)
        self.assertDictEqual(jsonld_dict,
                             ref_jsonld)

        jsonld_dict = json.loads(mt.model_dump_jsonld())
        jsonld_dict.pop('@id', None)
        self.assertDictEqual(jsonld_dict,
                             ref_jsonld)

    def test_blank_id_generator(self):
        @namespaces(foaf='http://xmlns.com/foaf/0.1/',
                    prov='https://www.w3.org/ns/prov#')
        @urirefs(Person='prov:Person',
                 first_name='foaf:firstName')
        class Person(Thing):
            first_name: str = Field(default=None, alias='firstName')

        counter = count()

        def my_generator():
            return f"_:{next(counter)}"

        with set_config(blank_id_generator=my_generator):
            p = Person(firstName="John")
            self.assertEqual("_:0", p.id)
            p = Person(firstName="John")
            self.assertEqual("_:1", p.id)

    def test_blank_node_prefix(self):
        @namespaces(foaf='http://xmlns.com/foaf/0.1/',
                    prov='https://www.w3.org/ns/prov#')
        @urirefs(Person='prov:Person',
                 first_name='foaf:firstName')
        class Person(Thing):
            first_name: str = Field(default=None, alias='firstName')

        p = Person(firstName="John")
        self.assertTrue(p.id.startswith("_:"))

        p = Person(firstName="John")
        self.assertTrue(p.id.startswith("_:"))

    def test_dynamic_thing(self):
        MyThing = build(
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="MyThing",
            properties=[Property(
                name="about",
                default=None,
                property_type=str
            )]
        )
        mything = MyThing(about="my thing")
        self.assertEqual(mything.about, "my thing")

    def test_dynamic_thing_with_validator(self):
        def validate_str(cls, value):
            if "thing" not in value:
                raise ValueError("Value must contain 'thing'")
            return value

        PositiveInt = Annotated[int, Field(gt=0)]

        MySpecialThing = build(
            baseclass=Thing,
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="MyThing",
            properties=[Property(
                name="value",
                default=Field(default=None, alias="val"),
                property_type=PositiveInt
            )]
        )
        with self.assertRaises(ValidationError):
            MySpecialThing(value=-1)
        mst = MySpecialThing(value=3)
        self.assertEqual(mst.value, 3)
        mst = MySpecialThing(val=4)
        self.assertEqual(mst.value, 4)

        MySpecialThing2 = Thing.build(
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="MyThing",
            properties=[Property(
                name="value",
                default=Field(default=None, alias="val"),
                property_type=PositiveInt
            )]
        )
        with self.assertRaises(ValidationError):
            MySpecialThing2(value=-1)
        mst = MySpecialThing2(value=3)
        self.assertEqual(mst.value, 3)
        mst = MySpecialThing2(val=4)
        self.assertEqual(mst.value, 4)

        NegativeInt = Annotated[int, Field(lt=0)]
        MySubSpecialThing = MySpecialThing.build(
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="MySubSpecialThing",
            properties=[Property(
                name="negValue",
                default=Field(default=None, alias="nval"),
                property_type=NegativeInt
            )]
        )
        with self.assertRaises(ValidationError):
            MySubSpecialThing(negValue=1)
        self.assertIsInstance(MySubSpecialThing(negValue=-1), MySubSpecialThing)
        mst = MySubSpecialThing(negValue=-3)
        self.assertEqual(mst.negValue, -3)
        self.assertEqual(mst.value, None)
        mst = MySubSpecialThing(nval=-4)
        self.assertEqual(mst.negValue, -4)

    def test_dynamic_forward_references(self):
        Person = Thing.build(
            namespace="https://example.org/",
            namespace_prefix="ex",
            class_name="Person",
            properties=[
                Property(
                    name="name",
                    default=None,
                    property_type=str
                ),
                Property(
                    name="friends",
                    default=None,
                    property_type=List["Person"]
                )
            ]
        )
        person1 = Person(name="John")
        person2 = Person(name="Jane", friends=[person1])
        expected_serialization = """@prefix ex: <https://example.org/> .

[] a ex:Person ;
    ex:friends [ a ex:Person ;
            ex:name "John" ] ;
    ex:name "Jane" .

"""
        self.assertEqual(expected_serialization, person2.serialize("ttl"))

    def test_skos_fields(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 name='foaf:lastName',
                 age='foaf:age')
        class Agent(Thing):
            name: str = Field(default=None, alias="lastName")  # name is synonymous to lastName
            age: int = None
            special_field: Optional[str] = None

        a = Agent(name='John Doe', age=23)
        b = Agent(name='John Doe', age=23)
        a.exactMatch = b
        expected_ttl = """@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a foaf:Agent ;
    skos:exactMatch [ a foaf:Agent ;
            foaf:age 23 ;
            foaf:lastName "John Doe" ] ;
    foaf:age 23 ;
    foaf:lastName "John Doe" .

"""
        self.assertEqual(expected_ttl, a.serialize(format="ttl"))

        a = Agent(id="https://example.org/jd", name='John Doe', age=23)
        b = Agent(name='John Doe', age=23, closeMatch=a.id)

        expected_ttl = """@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix skos: <http://www.w3.org/2004/02/skos/core#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a foaf:Agent ;
    skos:closeMatch <https://example.org/jd> ;
    foaf:age 23 ;
    foaf:lastName "John Doe" .

"""
        self.assertEqual(expected_ttl, b.serialize(format="ttl"))

    def test_behaviour_of_non_specified_fields(self):
        # if a Thing has not specified a specific field, but a user sets it, it cannot be validated, however,
        # if the user somehow passes the URI of the field, it should be accepted.

        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 name='foaf:lastName',
                 age='foaf:age')
        class Agent(Thing):
            name: str = Field(default=None, alias="lastName")  # name is synonymous to lastName
            age: int = None

        # in the following, home_town is not specified in the Agent class, but we set it anyway
        a = Agent(name='John Doe', age=23, homeTown=ontolutils.URIValue("Berlin", "http://example.org", "ex"))

    def test_relation(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 name='foaf:lastName',
                 age='foaf:age')
        class Agent(Thing):
            name: str = Field(default=None, alias="lastName")  # name is synonymous to lastName
            age: int = None
            special_field: Optional[str] = None

        a = Agent(name='John Doe', age=23, relation="https://example.org/123")

        expected_ttl = """@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a foaf:Agent ;
    dcterms:relation <https://example.org/123> ;
    foaf:age 23 ;
    foaf:lastName "John Doe" .

"""
        with self.assertRaises(ValueError):
            a = Agent(id="agents/123", name='John Doe', age=23, relation="https://example.org/123")

        a = Agent(id="_:agents/123", name='John Doe', age=23, relation="https://example.org/123")
        self.assertEqual(expected_ttl, a.serialize(format="ttl"))

        expected_ttl_with_base_uri = """@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

<https://example.org/agents/123> a foaf:Agent ;
    dcterms:relation <https://example.org/123> ;
    foaf:age 23 ;
    foaf:lastName "John Doe" .

"""
        self.assertEqual(expected_ttl_with_base_uri, a.serialize(format="ttl", base_uri="https://example.org/"))

        expected_ttl_with_base_uri2 = """@prefix agents: <https://example.org/agents/> .
@prefix dcterms: <http://purl.org/dc/terms/> .
@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

agents:123 a foaf:Agent ;
    dcterms:relation <https://example.org/123> ;
    foaf:age 23 ;
    foaf:lastName "John Doe" .

"""
        self.assertEqual(
            expected_ttl_with_base_uri2,
            a.serialize(format="ttl", base_uri="https://example.org/",
                        context={"agents": "https://example.org/agents/"})
        )

    def test_different_python_classes_same_uri(self):
        @namespaces(foaf='http://xmlns.com/foaf/0.1/',
                    prov='https://www.w3.org/ns/prov#')
        @urirefs(Person1='prov:Person',
                 first_name='foaf:firstName')
        class Person1(Thing):
            first_name: str = Field(default=None, alias='firstName')

        p1 = Person1(first_name='John')
        self.assertEqual(p1.namespace, "https://www.w3.org/ns/prov#")
        self.assertEqual(p1.uri, "https://www.w3.org/ns/prov#Person")

        @namespaces(foaf='http://xmlns.com/foaf/0.1/',
                    prov='https://www.w3.org/ns/prov#')
        @urirefs(Person2='prov:Person',
                 first_name='foaf:firstName',
                 last_name='foaf:lastName')
        class Person2(Thing):
            first_name: str = Field(alias='firstName')
            last_name: str = Field(alias='lastName')

        p1 = Person1(first_name='John')
        p2 = Person2(first_name='John', last_name="Doe")
        self.assertNotEqual(p1, p2)
        self.assertEqual(p1.namespace, p2.namespace)

        p1_from_p2 = p2.map(Person1)
        self.assertIsInstance(p1_from_p2, Thing)
        self.assertIsInstance(p1_from_p2, Person1)
        self.assertEqual(p1_from_p2.first_name, 'John')
        self.assertEqual(p1_from_p2.last_name, 'Doe')
        jsonld_dict = json.loads(p1_from_p2.model_dump_jsonld())
        self.assertTrue("foaf:lastName" in jsonld_dict)

        @namespaces(ex='http://example.org/',
                    prov='https://www.w3.org/ns/prov#')
        @urirefs(Organization='ex:Organization',
                 members='prov:Person')
        class Organization(Thing):
            members: Union[Person1, List[Person1]]

        # it should be irrelevant which Person class is taken, unless the uri is different
        org1 = Organization(members=p1)
        org2 = Organization(members=p2.map(Person1))
        # org3 = Organization(members=[p1, p2])

    def test_about(self):
        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 name='foaf:lastName')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            name: str = Field(default=None, alias="lastName")  # name is synonymous to lastName

        #         with self.assertRaises(ValidationError):
        #             Agent(
        #                 name="John Doe",
        #                 about="A person"
        #             )
        #         p = Agent(
        #             name="John Doe",
        #             about="http://example.org/123"
        #         )
        #         p = Agent(
        #             name="John Doe",
        #             about=["http://example.org/123", "http://example.org/456"]
        #         )
        #         p = Agent(
        #             name="John Doe",
        #             about=["http://example.org/123", Thing(id="http://example.org/456")]
        #         )
        #         ttl = p.model_dump_ttl()
        #         self.assertEqual(ttl, """@prefix foaf: <http://xmlns.com/foaf/0.1/> .
        # @prefix owl: <http://www.w3.org/2002/07/owl#> .
        # @prefix schema: <https://schema.org/> .
        #
        # <http://example.org/456> a owl:Thing .
        #
        # [] a foaf:Agent ;
        #     foaf:lastName "John Doe" ;
        #     schema:about <http://example.org/456>,
        #         "http://example.org/123" .
        #
        # """)
        p = Agent(
            name="John Doe",
            about=SCHEMA.about
        )
        ttl = p.model_dump_ttl()
        print(ttl)
        self.assertEqual(ttl, """@prefix foaf: <http://xmlns.com/foaf/0.1/> .
@prefix schema: <https://schema.org/> .

[] a foaf:Agent ;
    foaf:lastName "John Doe" ;
    schema:about schema:about .

""")



import pathlib
import unittest

__this_dir__ = pathlib.Path(__file__).parent

from ontolutils.ex import skos


class TestSkos(unittest.TestCase):

    def test_concept(self):
        concept = skos.Concept(
            id="http://example.org/concept/1",
            prefLabel="Example Concept@en",
        )
        self.assertEqual(
            concept.serialize("ttl"),
            """@prefix skos: <http://www.w3.org/2004/02/skos/core#> .

<http://example.org/concept/1> a skos:Concept ;
    skos:prefLabel "Example Concept"@en .

"""
        )

    def test_ConceptScheme(self):
        concept = skos.Concept(
            id="http://example.org/concept/1",
            prefLabel="Example Concept@en",
        )
        scheme = skos.ConceptScheme(
            id="http://example.org/scheme/1",
            prefLabel="Example Scheme@en",
            has_top_concept=[concept],
        )
        self.assertEqual(scheme.serialize("ttl"),
                         """@prefix skos: <http://www.w3.org/2004/02/skos/core#> .

<http://example.org/scheme/1> a skos:ConceptScheme ;
    skos:hasTopConcept <http://example.org/concept/1> .

<http://example.org/concept/1> a skos:Concept ;
    skos:prefLabel "Example Concept"@en .

""")



"""Testing code used in the README.md file"""
import json
import unittest
from typing import List, Union

import rdflib
from pydantic import EmailStr
from pydantic import Field
from pydantic import HttpUrl, model_validator

from ontolutils import Thing, urirefs, namespaces
from ontolutils import as_id
from ontolutils import build, Property


class TestReadmeCode(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None

    def test_code1_on_readme(self):
        """Just has to run without errors"""

        @namespaces(prov="https://www.w3.org/ns/prov#",
                    foaf="http://xmlns.com/foaf/0.1/",
                    m4i='http://w3id.org/nfdi4ing/metadata4ing#')
        @urirefs(Person='prov:Person',
                 firstName='foaf:firstName',
                 lastName='foaf:lastName',
                 mbox='foaf:mbox',
                 orcidId='m4i:orcidId')
        class Person(Thing):
            firstName: str
            lastName: str = Field(default=None, alias="last_name")  # you may provide an alias
            mbox: EmailStr = Field(default=None, alias="email")
            orcidId: HttpUrl = Field(default=None, alias="orcid_id")

            # the following will ensure, that if orcidId is set, it will be used as the id
            @model_validator(mode="before")
            def _change_id(self):
                return as_id(self, "orcidId")

        p = Person(
            id="https://orcid.org/0000-0001-8729-0482",
            label=rdflib.Literal("The creator of this package", lang="en"),
            firstName='Matthias',
            last_name='Probst'
        )
        # as we have set an alias, we can also use "lastName":
        p = Person(
            id="https://orcid.org/0000-0001-8729-0482",
            label=rdflib.Literal("The creator of this package", lang="en"),
            firstName='Matthias',
            lastName='Probst'
        )

        json_ld_serialization = p.model_dump_jsonld()
        serialized_str = p.serialize(format="json-ld")
        expected = """{
  "@context": {
    "owl": "http://www.w3.org/2002/07/owl#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "dcterms": "http://purl.org/dc/terms/",
    "schema": "https://schema.org/",
    "skos": "http://www.w3.org/2004/02/skos/core#",
    "prov": "https://www.w3.org/ns/prov#",
    "foaf": "http://xmlns.com/foaf/0.1/",
    "m4i": "http://w3id.org/nfdi4ing/metadata4ing#"
  },
  "@id": "https://orcid.org/0000-0001-8729-0482",
  "@type": "prov:Person",
  "rdfs:label": {"@language": "en", "@value": "The creator of this package"},
  "foaf:firstName": "Matthias",
  "foaf:lastName": "Probst"
}"""
        self.assertDictEqual(json.loads(json_ld_serialization), json.loads(expected))
        self.assertDictEqual(json.loads(serialized_str), json.loads(expected))

    def test_code2_on_readme(self):
        Event = build(
            namespace="https://schema.org/",
            namespace_prefix="schema",
            class_name="Event",
            properties=[Property(
                name="about",
                default=None,
                property_type=Union[Thing, List[Thing]]
            )]
        )
        conference = Event(label="my conference", about=[Thing(label='The thing it is about')])
        ttl = conference.serialize(format="ttl")
        expected = """@prefix owl: <http://www.w3.org/2002/07/owl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix schema: <https://schema.org/> .

[] a schema:Event ;
    rdfs:label "my conference" ;
    schema:about [ a owl:Thing ;
            rdfs:label "The thing it is about" ] .

"""
        self.assertEqual(ttl, expected)

    def test_person_with_base_uri(self):
        @namespaces(prov="http://www.w3.org/ns/prov#",
                    foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Person='prov:Person',
                 firstName='foaf:firstName',
                 lastName='foaf:lastName',
                 mbox='foaf:mbox')
        class Person(Thing):
            firstName: str
            lastName: str = None
            mbox: EmailStr = None

        person = Person(id='_:123uf4', label='test_person', firstName="John", mbox="john@email.com")
        json_ld_str = person.model_dump_jsonld(
            base_uri="https://example.org/",
            context={"ex": "https://example.org/"},
            resolve_keys=True
        )
        json_ld_dict = json.loads(json_ld_str)
        expected_dict = json.loads("""{
    "@context": {
        "owl": "http://www.w3.org/2002/07/owl#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "dcterms": "http://purl.org/dc/terms/",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "prov": "http://www.w3.org/ns/prov#",
        "foaf": "http://xmlns.com/foaf/0.1/",
        "schema": "https://schema.org/",
        "ex": "https://example.org/"
    },
    "@type": "prov:Person",
    "rdfs:label": "test_person",
    "foaf:firstName": "John",
    "foaf:mbox": "john@email.com",
    "@id": "ex:123uf4"
}""")
        self.assertEqual(
            expected_dict,
            json_ld_dict
        )



"""

utility for developers to automatically write namespaces python files automatically

This module should not be called during normal operation. It is only for developers who want to
update the namespace files.

"""
import datetime
import json
import pathlib
import warnings
from typing import Iterable, Dict, Union, Optional

import requests
from rdflib import Graph, BNode
from rdflib import OWL, RDF, RDFS

from ontolutils import __version__

__this_dir__ = pathlib.Path(__file__).parent
__package_dir__ = __this_dir__.parent / 'namespacelib'

FORCE_DOWNLOAD = True

FORBIDDEN_PROPERTIES = ["and", "or", "type", "yield", "True", "False", "in", "not", "is", "as", "if", "else", "elif", ]


def generate_namespace_file_from_ttl(namespace: str,
                                     source: str,
                                     ns: str,
                                     is_owl: bool = False,
                                     target_dir: Optional[Union[str, pathlib.Path]] = None,
                                     fail=True):
    """Generate M4I_NAMESPACE.py file from m4i_context.jsonld
    """
    cls_type = RDFS.Class if not is_owl else OWL.Class
    prop_type = RDF.Property if not is_owl else OWL.ObjectProperty
    if is_owl:
        data_prop = OWL.DatatypeProperty
        named_individual = OWL.NamedIndividual
    else:
        named_individual = None
        data_prop = None

    g = Graph()
    g.parse(source)
    g.namespace_manager.bind(namespace, ns)
    if target_dir is None:
        target_dir = __this_dir__
    else:
        target_dir = pathlib.Path(target_dir)

    with open(target_dir / f'{namespace}.py', 'w',
              encoding='UTF8') as f:
        f.write('from rdflib.namespace import DefinedNamespace, Namespace\n')
        f.write('from rdflib.term import URIRef\n\n\n')
        f.write(f'class {namespace.upper()}(DefinedNamespace):')
        f.write(f'\n    # Generated with ontolutils version {__version__}')
        f.write(f'\n    _fail = {fail}')

        for cls in g.subjects(RDF.type, cls_type):
            # u = str(s).rsplit('/', 1)[-1].replace('-', '_')
            if not isinstance(cls, BNode):
                class_split = str(cls).split(ns)
                if len(class_split) == 2:
                    _key = class_split[-1]
                    if _key[0].isdigit():
                        _key = f'_{_key}'
                    f.write(f'\n    {_key}: URIRef')
        for prop in g.subjects(RDF.type, prop_type):
            if not isinstance(prop, BNode):
                prop_split = str(prop).split(ns)
                if len(prop_split) == 2:
                    if prop_split[-1] not in FORBIDDEN_PROPERTIES:
                        _key = prop_split[-1]
                        if _key[0].isdigit():
                            _key = f'_{_key}'
                        f.write(f'\n    {_key}: URIRef')

        if data_prop:
            for prop in g.subjects(RDF.type, data_prop):
                if not isinstance(prop, BNode):
                    prop_split = str(prop).split(ns)
                    if len(prop_split) == 2:
                        if prop_split[-1] not in FORBIDDEN_PROPERTIES:
                            f.write(f'\n    {prop_split[-1]}: URIRef')

        if named_individual:
            for prop in g.subjects(RDF.type, named_individual):
                if not isinstance(prop, BNode):
                    prop_split = str(prop).split(ns)
                    if len(prop_split) == 2:
                        if prop_split[-1] not in FORBIDDEN_PROPERTIES:
                            f.write(f'\n    {prop_split[-1]}: URIRef')

        f.write(f'\n\n    _NS = Namespace("{ns}")')


def generate_namespace_file_from_context(namespace: str,
                                         context_ld: str,
                                         languages: Optional[Dict[str, Iterable[str]]] = None,
                                         target_dir: Optional[Union[str, pathlib.Path]] = None,
                                         fail: bool = True,
                                         filename: Optional[Union[str, pathlib.Path]] = None):
    """Generate M4I_NAMESPACE.py file from m4i_context.jsonld"""
    languages = languages or {}
    assert isinstance(languages, dict)
    context_file = __this_dir__ / f'{namespace}.jsonld'

    if not context_file.exists() or FORCE_DOWNLOAD:
        with open(context_file, 'w', encoding='utf-8') as f:
            f.write(requests.get(context_ld).text, )

    # read context file:
    with open(context_file, encoding='utf-8') as f:
        context = json.load(f)

    url = context['@context'][namespace]

    iris = {}
    for k, v in context['@context'].items():
        if k not in ('type', 'id'):
            if '@id' in v:
                if isinstance(v, dict) and namespace in v['@id']:
                    name = v["@id"].rsplit(":", 1)[-1]

                    if '#' in name:
                        warnings.warn(f'Skipping {name} ({v}) because it has a "#" in it.')
                    elif name[0].isdigit():
                        warnings.warn(f'Skipping {name} ({v}) because it starts with a digit.')
                    elif name in ('True', 'False', 'yield'):
                        warnings.warn(f'Skipping {name} ({v}) because it starts with "yield".')
                    else:
                        if name not in iris:
                            iris[name] = {'url': f'{url}{name}', 'keys': [k, ]}
                        else:
                            iris[name]['keys'].append(k)
    if target_dir is None:
        target_dir = __this_dir__
    else:
        target_dir = pathlib.Path(target_dir)

    fields = []

    if filename is None:
        filename = target_dir / f'{namespace}.py'
    else:
        filename = pathlib.Path(filename)
    with open(filename, 'w',
              encoding='UTF8') as f:
        f.write('from rdflib.namespace import DefinedNamespace, Namespace\n')
        f.write('from rdflib.term import URIRef\n')
        f.write('\n\nclass LanguageExtension:\n    pass')
        f.write(f'\n\nclass {namespace.upper()}(DefinedNamespace):')
        f.write(f'\n    # uri = "{url}"')
        f.write(f'\n    # Generated with ontolutils version {__version__}')
        f.write(f'\n    # Date: {datetime.datetime.now()}')
        f.write(f'\n    _fail = {fail}')
        for k, v in iris.items():
            if k not in fields:
                fields.append(k)
                if '-' in k:
                    f.write(f'\n    {k.replace("-", "_")} = URIRef("{url}{k}")  # {v["keys"]}')
                else:
                    f.write(f'\n    {k}: URIRef  # {v["keys"]}')

        f.write(f'\n\n    _NS = Namespace("{url}")')

        for lang in languages:
            f.write(f'\n\n{lang} = LanguageExtension()')

        f.write('\n')

        for k, v in iris.items():
            for kk in v["keys"]:
                found_language_key = False
                key = kk.replace(' ', '_')
                for lang_key, lang_values in languages.items():
                    if key in lang_values:
                        f.write(f'\nsetattr({lang_key}, "{key}", {namespace.upper()}.{k.replace("-", "_")})')
                        found_language_key = True
                if not found_language_key:
                    f.write(f'\nsetattr({namespace.upper()}, "{key}", {namespace.upper()}.{k.replace("-", "_")})')

        for lang in languages:
            f.write(f'\n\nsetattr({namespace.upper()}, "{lang}", {lang})')
    context_file.unlink(missing_ok=True)


def generate_namespace_file_from_owl_xml(
        namespace: str,
        xml_source: str,
        target_dir: Optional[Union[str, pathlib.Path]] = None,
        fail: bool = True,
        filename: Optional[Union[str, pathlib.Path]] = None,
        ns: str = None,
):
    xml_file = __this_dir__ / f'{namespace}.xml'

    if not xml_file.exists() or FORCE_DOWNLOAD:
        with open(xml_file, 'w', encoding='utf-8') as f:
            f.write(requests.get(xml_source).text, )
    g = Graph()
    g.parse(xml_file, format="xml")

    if target_dir is None:
        target_dir = __this_dir__
    else:
        target_dir = pathlib.Path(target_dir)
    if filename is None:
        filename = target_dir / f'{namespace}.py'
    else:
        filename = pathlib.Path(filename)
    with open(filename, 'w', encoding='UTF8') as f:
        f.write('from rdflib.namespace import DefinedNamespace, Namespace\n')
        f.write('from rdflib.term import URIRef\n\n\n')
        f.write(f'class {namespace.upper()}(DefinedNamespace):')
        f.write(f'\n    _fail = {fail}')

        # Klassen
        for cls in g.subjects(RDF.type, OWL.Class):
            if not isinstance(cls, BNode):
                class_split = str(cls).split(ns)
                if len(class_split) == 2:
                    _key = class_split[-1]
                    if _key[0].isdigit():
                        _key = f'_{_key}'
                    f.write(f'\n    {_key}: URIRef')

        # Objekt-Properties
        for prop in g.subjects(RDF.type, OWL.ObjectProperty):
            if not isinstance(prop, BNode):
                prop_split = str(prop).split(ns)
                if len(prop_split) == 2 and prop_split[-1] not in FORBIDDEN_PROPERTIES:
                    _key = prop_split[-1]
                    if _key[0].isdigit():
                        _key = f'_{_key}'
                    f.write(f'\n    {_key}: URIRef')

        # Daten-Properties
        for prop in g.subjects(RDF.type, OWL.DatatypeProperty):
            if not isinstance(prop, BNode):
                prop_split = str(prop).split(ns)
                if len(prop_split) == 2 and prop_split[-1] not in FORBIDDEN_PROPERTIES:
                    _key = prop_split[-1]
                    if _key[0].isdigit():
                        _key = f'_{_key}'
                    f.write(f'\n    {_key}: URIRef')

        # Named Individuals
        for ind in g.subjects(RDF.type, OWL.NamedIndividual):
            if not isinstance(ind, BNode):
                ind_split = str(ind).split(ns)
                if len(ind_split) == 2:
                    _key = ind_split[-1]
                    if _key[0].isdigit():
                        _key = f'_{_key}'
                    f.write(f'\n    {_key}: URIRef')

        f.write(f'\n\n    _NS = Namespace("{ns}")')


def m4i():
    """generates namespace for metadata4Ing ontology (M4I)"""
    generate_namespace_file_from_context(
        'm4i',
        context_ld='http://w3id.org/nfdi4ing/metadata4ing/m4i_context.jsonld',
        languages={'de': [
            'Methode',
            'numersiche_Zuweisung',
            'numerische_Variable',
            'Arbeitsschritt',
            'textbasierte_Variable',
            'Werkzeug',
            'Unsicherheitsdeklaration',
            'hat_als_zulässige_Einheit',
            'hat_als_zulässigen_Wert',
            'hat_zugewiesenen_Wert',
            'hat_Überdeckungsintervall',
            'hat_eingesetztes_Werkzeug',
            'hat_erweiterte_Unsicherheit',
            'hat_Größenart',
            'hat_Parameter',
            'hat_Laufzeitzuweisung',
            'hat_Unsicherheitsdeklaration',
            'hat_Einheit',
            'hat_Variable',
            'gehört_zu_Projekt',
            'untersucht',
            'untersucht_Eigenschaft',
            'ist_eingesetztes_Werkzeug',
            'hat_Projektmitglied',
            'realisiert_Methode',
            'Verwendungshinweis',
            'Projektenddatum',
            'hat_Zuweisungszeitstempel',
            'hat_Datumszuweisung_erzeugt',
            'hat_Datumszuweisung_gelöscht',
            'hat_Datumszuweisung_bearbeitet',
            'hat_Datumszuweisung_gültig_ab',
            'hat_Datumszuweisung_gültig_bis',
            'hat_Maximalwert',
            'hat_Minimalwert',
            'hat_Zahlenwert',
            'hat_Schrittweite',
            'hat_Zeichenwert',
            'hat_Symbol',
            'hat_Wert',
            'hat_Variablenbeschreibung',
            'hat_Identifikator',
            'hat_ORCID_ID',
            'hat_Projekt-ID',
            'Projektstartdatum',
            'Kontaktperson',
            'Datenerfasser*in',
            'Datenkurator*in',
            'Datenkurator*in',
            'Datenverwalter*in',
            'Anbieter*in',
            'Herausgeber*in',
            'bereitstellende_Institution',
            'weitere_Person',
            'Produzent*in',
            'Projektleiter*in',
            'Projektmanager*in',
            'Projektmitglied',
            'Registrierungsstelle',
            'Registrierungsbehörde',
            'zugehörige_Person',
            'Forschungsgruppe',
            'Rechercheur*in',
            'Rechteinhaber*in',
            'Sponsor*in',
            'Betreuer*in',
            'Arbeitspaketleiter*in',
        ]}
    )


def spdx():
    """generates namespace for The System Package Data Exchange ontology (spdx)"""
    generate_namespace_file_from_owl_xml(
        "spdx",
        xml_source="https://raw.githubusercontent.com/spdx/spdx-spec/refs/heads/development/v2.2.2/ontology/spdx-ontology.owl.xml",
        ns="http://spdx.org/rdf/terms#",
    )


def obo():
    """Generate obo ontology namespace class"""
    generate_namespace_file_from_context(
        'obo',
        # context_ld='https://raw.githubusercontent.com/biopragmatics/bioregistry/main/exports/contexts/obo.context.jsonld',
        context_ld='http://w3id.org/nfdi4ing/metadata4ing/m4i_context.jsonld',
        languages={'de': ['Prozess',
                          'realisierbare_Entität',
                          'Teil_von',
                          'hat_Teil',
                          'realisiert_in',
                          'realisiert',
                          'ist_Voraussetzung_für_Schritt',
                          'ist_beteiligt_an',
                          'hat_Teilnehmer',
                          'ist_unmittelbare_Voraussetzung_für_Schritt',
                          'beginnt_mit',
                          'endet_mit',
                          'hat_Input',
                          'hat_Output',
                          'Input_von',
                          'Output_von',

                          ]
                   }
    )
    # # be careful, german lines must be manually uncommented
    # # generate_qudt_unit_namespace()  # write _qudt_namespace.py manually
    # generate_qudt_quantitykind_namespace()  # write _qudt_quantitykind_namespace.py manually
    # generate_codemeta_namespace()


def qudt_unit():
    generate_namespace_file_from_ttl(
        namespace='qudt_unit',
        source='http://qudt.org/vocab/unit/',
        ns='http://qudt.org/vocab/unit/',
    )


def hdf5():
    generate_namespace_file_from_ttl(
        namespace='hdf5',
        source='https://purl.allotrope.org/voc/adf/REC/2024/12/hdf.ttl',
        ns='http://purl.allotrope.org/ontologies/hdf5/1.8#',
        is_owl=True
    )


def qudt_quantitykind():
    generate_namespace_file_from_ttl(
        namespace='qudt_kind',
        source='http://qudt.org/vocab/quantitykind/',
        ns='http://qudt.org/vocab/quantitykind/',
    )


def schema():
    generate_namespace_file_from_ttl(
        namespace='schema',
        source='https://schema.org/version/latest/schemaorg-current-https.jsonld',
        ns='https://schema.org/',
    )


def codemeta():
    generate_namespace_file_from_context(
        namespace='codemeta',
        context_ld='https://raw.githubusercontent.com/codemeta/codemeta/2.0/codemeta.jsonld'
    )


def build_namespace_files():
    """Call this only if you are a developer and want to build the namespace files"""
    # with open(__package_dir__ / '__init__.py', 'w') as f:
    #     f.write('"""Auto-generated file. Do not edit!"""\n')
    # f.write('from ._version import __version__\n')

    # m4i()
    #
    # obo()
    #
    # qudt_unit()
    #
    # qudt_quantitykind()
    #
    # codemeta()
    #
    # schema()

    # hdf5()
    spdx()
    # with open(__package_dir__ / '__init__.py', 'a') as f:
    # f.write('from .m4i import M4I\n')
    # f.write('from .obo import OBO\n')
    # f.write('from .pivmeta import PIVMETA\n')
    # f.write('from .codemeta import CODEMETA\n')
    # f.write('from .qudt_unit import QUDT_UNIT\n')
    # f.write('from .qudt_kind import QUDT_KIND\n')
    # f.write('from .schema import SCHEMA\n')
    # f.write('from .ssno import SSNO\n')
    # f.write('from ._iana_utils import IANA\n')

    for jld in __package_dir__.glob('*.jsonld'):
        jld.unlink()


if __name__ == '__main__':
    """Call this only if you are a developer and want to build the namespace files"""
    build_namespace_files()



from datetime import datetime
from typing import Optional, List, Union

from pydantic import Field

from ontolutils import Thing, LangString, namespaces, urirefs
from ontolutils.typing import ResourceType


@namespaces(skos="http://www.w3.org/2004/02/skos/core#",
            dcterms="http://purl.org/dc/terms/",
            prov="http://www.w3.org/ns/prov#")
@urirefs(Note='skos:Note',
         value='rdf:value',
         created='dcterms:created',
         wasRevisionOf='prov:wasRevisionOf',
         creator='dcterms:creator')
class Note(Thing):
    value: LangString = Field(...)
    created: Optional[datetime] = Field(default=None)
    wasRevisionOf: Optional[Thing] = Field(default=None, alias="wasRevisionOf")
    creator: Optional[Union[ResourceType, List[ResourceType]]] = Field(default=None)


@namespaces(skos="http://www.w3.org/2004/02/skos/core#")
@urirefs(ConceptScheme='skos:ConceptScheme',
         hasTopConcept='skos:hasTopConcept'
         )
class ConceptScheme(Thing):
    """Implementation of skos:ConceptScheme"""
    hasTopConcept: Optional[Union['Concept', List['Concept']]] = Field(default=None, alias="has_top_concept")


@namespaces(skos="http://www.w3.org/2004/02/skos/core#")
@urirefs(Concept='skos:Concept',
         prefLabel='skos:prefLabel',
         altLabel='skos:altLabel',
         hiddenLabel='skos:hiddenLabel',
         definition='skos:definition',
         note='skos:note',
         scopeNote='skos:scopeNote',
         editorialNote='skos:editorialNote',
         changeNote='skos:changeNote',
         inScheme='skos:inScheme',
         example='skos:example')
class Concept(Thing):
    """Implementation of skos:Concept"""

    prefLabel: Optional[Union[LangString, List[LangString]]] = Field(default=None, alias="pref_label")
    altLabel: Optional[Union[LangString, List[LangString]]] = Field(default=None, alias="alt_label")
    hiddenLabel: Optional[Union[LangString, List[LangString]]] = Field(default=None, alias="hidden_label")
    definition: Optional[Union[LangString, List[LangString]]] = Field(default=None)
    note: Optional[Union[LangString, Note, List[Union[LangString, Note]]]] = Field(default=None)
    scopeNote: Optional[Union[LangString, Note, List[Union[LangString, Note]]]] = Field(default=None,
                                                                                        alias="scope_note")
    editorialNote: Optional[Union[LangString, Note, List[Union[LangString, Note]]]] = Field(default=None,
                                                                                            alias="editorial_note")
    changeNote: Optional[Union[LangString, Note, List[Union[LangString, Note]]]] = Field(default=None,
                                                                                         alias="change_note")
    inScheme: Optional[Union[ConceptScheme, List[ConceptScheme]]] = Field(default=None, alias="in_scheme")
    example: Optional[Union[LangString, List[LangString]]] = Field(default=None)


ConceptScheme.model_rebuild()



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class HDF5(DefinedNamespace):
    # Generated with ontolutils version 0.13.2
    _fail = True
    AllSelection: URIRef
    AndSelection: URIRef
    ArrayDatatype: URIRef
    ArrayDimension: URIRef
    ArrayValue: URIRef
    AtomicDatatype: URIRef
    Attribute: URIRef
    BitfieldDatatype: URIRef
    CharacterSet: URIRef
    Chunk: URIRef
    ChunkDimension: URIRef
    ChunkMode: URIRef
    CompositeDatatype: URIRef
    CompositeValue: URIRef
    CompoundDatatype: URIRef
    CompoundDatatypeMember: URIRef
    CompoundValue: URIRef
    Coordinate: URIRef
    DataStream: URIRef
    Dataset: URIRef
    Dataspace: URIRef
    DataspaceDimension: URIRef
    Datatype: URIRef
    Dimension: URIRef
    DimensionSelection: URIRef
    EnumerationDatatype: URIRef
    EnumerationDatatypeMember: URIRef
    ErrorDetectionCheck: URIRef
    ExponentField: URIRef
    ExternalFile: URIRef
    FieldSpecification: URIRef
    File: URIRef
    FileAccessMode: URIRef
    FillTime: URIRef
    FillValueStatus: URIRef
    Filter: URIRef
    FilterDefault: URIRef
    FilterDeflate: URIRef
    FilterFletcher: URIRef
    FilterNBit: URIRef
    FilterSZip: URIRef
    FilterScaleOffset: URIRef
    FilterShuffle: URIRef
    FloatAttribute: URIRef
    FloatDatatype: URIRef
    FloatingPointNormalization: URIRef
    Group: URIRef
    HardLink: URIRef
    HyperSlab: URIRef
    IntegerAttribute: URIRef
    IntegerDatatype: URIRef
    Link: URIRef
    MantissaField: URIRef
    MessageTypeFlag: URIRef
    NameValuePair: URIRef
    NamedDatatype: URIRef
    NamedObject: URIRef
    NoneSelection: URIRef
    NotASelection: URIRef
    NotBSelection: URIRef
    OpaqueAttribute: URIRef
    OpaqueDatatype: URIRef
    OrSelection: URIRef
    Order: URIRef
    Padding: URIRef
    Point: URIRef
    PointSelection: URIRef
    PointValuePair: URIRef
    ReferenceDatatype: URIRef
    SZipCoding: URIRef
    ScalarDataspace: URIRef
    ScaleType: URIRef
    Selection: URIRef
    Sign: URIRef
    SignField: URIRef
    SimpleDataspace: URIRef
    SoftLink: URIRef
    StorageAllocation: URIRef
    StorageLayout: URIRef
    StringAttribute: URIRef
    StringDatatype: URIRef
    StringPadding: URIRef
    TimeDatatype: URIRef
    TransferMode: URIRef
    TypeClass: URIRef
    VariableLengthDatatype: URIRef
    XOrSelection: URIRef
    allocationTime: URIRef
    attribute: URIRef
    characterSet: URIRef
    chunk: URIRef
    chunkMode: URIRef
    compositeFillValue: URIRef
    compositeValue: URIRef
    coordinate: URIRef
    dataSource: URIRef
    dataTarget: URIRef
    dataspace: URIRef
    datatype: URIRef
    dimension: URIRef
    dimensionSelection: URIRef
    errorDetectionCheck: URIRef
    exponentField: URIRef
    externalLinkAccess: URIRef
    field: URIRef
    fillTime: URIRef
    fillValueStatus: URIRef
    filter: URIRef
    internalPadding: URIRef
    layout: URIRef
    link: URIRef
    linkedTo: URIRef
    lsbPadding: URIRef
    mantissaField: URIRef
    member: URIRef
    msbPadding: URIRef
    multiDimensionalIndex: URIRef
    normalization: URIRef
    notA: URIRef
    notB: URIRef
    order: URIRef
    point: URIRef
    previous: URIRef
    rootGroup: URIRef
    scaleType: URIRef
    selectionOn: URIRef
    sign: URIRef
    signField: URIRef
    stringPadding: URIRef
    szipEncoding: URIRef
    transferMode: URIRef
    typeClass: URIRef
    writtenRegion: URIRef
    xor: URIRef
    atomicFillValue: URIRef
    block: URIRef
    btreeIStoreSize: URIRef
    btreeLeaf: URIRef
    btreeRank: URIRef
    btreeRatioLeft: URIRef
    btreeRatioMiddle: URIRef
    btreeRatioRight: URIRef
    bufferSize: URIRef
    chunkCachePreemptionPolicy: URIRef
    chunkCacheSize: URIRef
    chunkCacheSlots: URIRef
    collectiveIOThreshold: URIRef
    compressionLevel: URIRef
    coordinateIndex: URIRef
    count: URIRef
    createIntermediateGroups: URIRef
    currentSize: URIRef
    data: URIRef
    deflateLevel: URIRef
    dimensionIndex: URIRef
    ebias: URIRef
    externalLinkCacheSize: URIRef
    externalLinkPrefix: URIRef
    globalFreelistVersion: URIRef
    hyperVectorSize: URIRef
    iStoreK: URIRef
    index: URIRef
    index_0: URIRef
    index_1: URIRef
    index_10: URIRef
    index_11: URIRef
    index_12: URIRef
    index_13: URIRef
    index_14: URIRef
    index_15: URIRef
    index_16: URIRef
    index_17: URIRef
    index_18: URIRef
    index_19: URIRef
    index_2: URIRef
    index_20: URIRef
    index_21: URIRef
    index_22: URIRef
    index_23: URIRef
    index_24: URIRef
    index_25: URIRef
    index_26: URIRef
    index_27: URIRef
    index_28: URIRef
    index_29: URIRef
    index_3: URIRef
    index_30: URIRef
    index_31: URIRef
    index_4: URIRef
    index_5: URIRef
    index_6: URIRef
    index_7: URIRef
    index_8: URIRef
    index_9: URIRef
    initialSize: URIRef
    isVariableLength: URIRef
    linkedChunkThreshold: URIRef
    maximumLinkTraversals: URIRef
    maximumSize: URIRef
    metaBlockSize: URIRef
    minimumSize: URIRef
    name: URIRef
    offset: URIRef
    pixelsPerBlock: URIRef
    position: URIRef
    precision: URIRef
    preserve: URIRef
    rank: URIRef
    scaleFactor: URIRef
    sharedObjectHeaderVersion: URIRef
    sieveBufferSize: URIRef
    size: URIRef
    sizeOfAddress: URIRef
    sizeOfSize: URIRef
    smallDataBlockSize: URIRef
    start: URIRef
    stride: URIRef
    superBlockVersion: URIRef
    symbolTableNodeSize: URIRef
    symbolTableTreeRank: URIRef
    symbolTableVersion: URIRef
    trackTimes: URIRef
    userBlockSize: URIRef
    value: URIRef
    H5D_ALLOC_TIME_DEFAULT: URIRef
    H5D_ALLOC_TIME_EARLY: URIRef
    H5D_ALLOC_TIME_INCR: URIRef
    H5D_ALLOC_TIME_LATE: URIRef
    H5D_CHUNKED: URIRef
    H5D_COMPACT: URIRef
    H5D_CONTIGUOUS: URIRef
    H5D_FILL_TIME_ALLOC: URIRef
    H5D_FILL_TIME_IFSET: URIRef
    H5D_FILL_TIME_NEVER: URIRef
    H5D_FILL_VALUE_DEFAULT: URIRef
    H5D_FILL_VALUE_UNDEFINED: URIRef
    H5D_FILL_VALUE_USER_DEFINED: URIRef
    H5D_VIRTUAL: URIRef
    H5FD_MPIO_CHUNK_MULTI_IO: URIRef
    H5FD_MPIO_CHUNK_ONE_IO: URIRef
    H5FD_MPIO_COLLECTIVE: URIRef
    H5FD_MPIO_INDEPENDENT: URIRef
    H5F_ACC_DEFAULT: URIRef
    H5F_ACC_RDONLY: URIRef
    H5F_ACC_RDWR: URIRef
    H5O_SHMESG_ALL_FLAG: URIRef
    H5O_SHMESG_ATTR_FLAG: URIRef
    H5O_SHMESG_DTYPE_FLAG: URIRef
    H5O_SHMESG_FILL_FLAG: URIRef
    H5O_SHMESG_NONE_FLAG: URIRef
    H5O_SHMESG_PLINE_FLAG: URIRef
    H5O_SHMESG_SDSPACE_FLAG: URIRef
    H5P_DATASET_CREATE: URIRef
    H5P_DATASET_XFER: URIRef
    H5P_FILE_ACCESS: URIRef
    H5P_FILE_CREATE: URIRef
    H5P_FILE_MOUNT: URIRef
    H5T_ALPHA_B16: URIRef
    H5T_ALPHA_B32: URIRef
    H5T_ALPHA_B64: URIRef
    H5T_ALPHA_B8: URIRef
    H5T_ALPHA_F32: URIRef
    H5T_ALPHA_F64: URIRef
    H5T_ALPHA_I16: URIRef
    H5T_ALPHA_I32: URIRef
    H5T_ALPHA_I64: URIRef
    H5T_ALPHA_I8: URIRef
    H5T_ALPHA_U16: URIRef
    H5T_ALPHA_U32: URIRef
    H5T_ALPHA_U64: URIRef
    H5T_ALPHA_U8: URIRef
    H5T_ARRAY: URIRef
    H5T_BITFIELD: URIRef
    H5T_COMPOUND: URIRef
    H5T_CSET_ASCII: URIRef
    H5T_CSET_UTF8: URIRef
    H5T_ENUM: URIRef
    H5T_FLOAT: URIRef
    H5T_IEEE_F32BE: URIRef
    H5T_IEEE_F32LE: URIRef
    H5T_IEEE_F64BE: URIRef
    H5T_IEEE_F64LE: URIRef
    H5T_INTEGER: URIRef
    H5T_INTEL_B16: URIRef
    H5T_INTEL_B32: URIRef
    H5T_INTEL_B64: URIRef
    H5T_INTEL_B8: URIRef
    H5T_INTEL_F32: URIRef
    H5T_INTEL_F64: URIRef
    H5T_INTEL_I16: URIRef
    H5T_INTEL_I32: URIRef
    H5T_INTEL_I64: URIRef
    H5T_INTEL_I8: URIRef
    H5T_INTEL_U16: URIRef
    H5T_INTEL_U32: URIRef
    H5T_INTEL_U64: URIRef
    H5T_INTEL_U8: URIRef
    H5T_MIPS_B16: URIRef
    H5T_MIPS_B32: URIRef
    H5T_MIPS_B64: URIRef
    H5T_MIPS_B8: URIRef
    H5T_MIPS_F32: URIRef
    H5T_MIPS_F64: URIRef
    H5T_MIPS_I16: URIRef
    H5T_MIPS_I32: URIRef
    H5T_MIPS_I64: URIRef
    H5T_MIPS_I8: URIRef
    H5T_MIPS_U16: URIRef
    H5T_MIPS_U32: URIRef
    H5T_MIPS_U64: URIRef
    H5T_MIPS_U8: URIRef
    H5T_NATIVE_CHAR: URIRef
    H5T_NATIVE_DOUBLE: URIRef
    H5T_NATIVE_FLOAT: URIRef
    H5T_NATIVE_INT: URIRef
    H5T_NATIVE_LDOUBLE: URIRef
    H5T_NATIVE_LLONG: URIRef
    H5T_NATIVE_LONG: URIRef
    H5T_NATIVE_SHORT: URIRef
    H5T_NATIVE_UCHAR: URIRef
    H5T_NATIVE_UINT: URIRef
    H5T_NATIVE_ULLONG: URIRef
    H5T_NATIVE_ULONG: URIRef
    H5T_NATIVE_USHORT: URIRef
    H5T_NORM_IMPLIED: URIRef
    H5T_NORM_MSBSET: URIRef
    H5T_NORM_NONE: URIRef
    H5T_OPAQUE: URIRef
    H5T_ORDER_BE: URIRef
    H5T_ORDER_LE: URIRef
    H5T_ORDER_MIXED: URIRef
    H5T_ORDER_VAX: URIRef
    H5T_PAD_BACKGROUND: URIRef
    H5T_PAD_ONE: URIRef
    H5T_PAD_ZERO: URIRef
    H5T_REFERENCE: URIRef
    H5T_SGN_2: URIRef
    H5T_SGN_NONE: URIRef
    H5T_STD_B16BE: URIRef
    H5T_STD_B16LE: URIRef
    H5T_STD_B32BE: URIRef
    H5T_STD_B32LE: URIRef
    H5T_STD_B64BE: URIRef
    H5T_STD_B64LE: URIRef
    H5T_STD_B8BE: URIRef
    H5T_STD_B8LE: URIRef
    H5T_STD_I16BE: URIRef
    H5T_STD_I16LE: URIRef
    H5T_STD_I32BE: URIRef
    H5T_STD_I32LE: URIRef
    H5T_STD_I64BE: URIRef
    H5T_STD_I64LE: URIRef
    H5T_STD_I8BE: URIRef
    H5T_STD_I8LE: URIRef
    H5T_STD_U16BE: URIRef
    H5T_STD_U16LE: URIRef
    H5T_STD_U32BE: URIRef
    H5T_STD_U32LE: URIRef
    H5T_STD_U64BE: URIRef
    H5T_STD_U64LE: URIRef
    H5T_STD_U8BE: URIRef
    H5T_STD_U8LE: URIRef
    H5T_STRING: URIRef
    H5T_STR_NULLPAD: URIRef
    H5T_STR_NULLTERM: URIRef
    H5T_STR_SPACEPAD: URIRef
    H5T_TIME: URIRef
    H5T_VAX_F32: URIRef
    H5T_VAX_F64: URIRef
    H5T_VLEN: URIRef
    H5Z_DISABLE_EDC: URIRef
    H5Z_ENABLE_EDC: URIRef
    H5Z_SO_FLOAT_DSCALE: URIRef
    H5Z_SO_FLOAT_ESCALE: URIRef
    H5Z_SO_INT: URIRef
    H5_SZIP_EC_OPTION_MASK: URIRef
    H5_SZIP_NN_OPTION_MASK: URIRef
    scalarDataspace: URIRef

    _NS = Namespace("http://purl.allotrope.org/ontologies/hdf5/1.8#")


import pathlib
import re
import shutil
from datetime import datetime
from typing import Union, List, Optional

from dateutil import parser
from ontolutils import Thing, as_id, urirefs, namespaces, LangString
from ontolutils.classes.utils import download_file
from ontolutils.ex import foaf
from ontolutils.ex import prov
from ontolutils.typing import BlankNodeType, ResourceType
from pydantic import HttpUrl, FileUrl, field_validator, Field, model_validator

from ..spdx import Checksum


@namespaces(dcat="http://www.w3.org/ns/dcat#",
            dcterms="http://purl.org/dc/terms/", )
@urirefs(Resource='dcat:Resource',
         title='dcterms:title',
         description='dcterms:description',
         creator='dcterms:creator',
         publisher='dcterms:publisher',
         contributor='dcterms:contributor',
         license='dcterms:license',
         version='dcat:version',
         identifier='dcterms:identifier',
         hasPart='dcterms:hasPart',
         keyword='dcat:keyword')
class Resource(Thing):
    """Pydantic implementation of dcat:Resource

    .. note::

        More than the below parameters are possible but not explicitly defined here.



    Parameters
    ----------
    title: str
        Title of the resource (dcterms:title)
    description: str = None
        Description of the resource (dcterms:description)
    creator: Union[
        foaf.Agent, foaf.Organization, foaf.Person, prov.Person, prov.Agent, prov.Organization, HttpUrl,
        List[Union[foaf.Agent, foaf.Organization, foaf.Person, prov.Person, prov.Agent, prov.Organization, HttpUrl]]
    ] = None
        Creator of the resource (dcterms:creator)
    publisher: Union[Agent, List[Agent]] = None
        Publisher of the resource (dcterms:publisher)
    contributor: Union[Agent, List[Agent]] = None
        Contributor of the resource (dcterms:contributor)
    license: ResourceType = None
        License of the resource (dcat:license)
    version: str = None
        Version of the resource (dcat:version),
        best following semantic versioning (https://semver.org/lang/de/)
    identifier: str = None
        Identifier of the resource (dcterms:identifier)
    hasPart: ResourceType = None
        A related resource that is included either physically or logically in the described resource. (dcterms:hasPart)
    keyword: List[str]
        Keywords for the distribution.
    """
    title: Optional[Union[LangString, List[LangString]]] = None  # dcterms:title
    description: Optional[Union[LangString, List[LangString]]] = None  # dcterms:description
    creator: Union[
        foaf.Agent,
        foaf.Organization,
        foaf.Person,
        prov.Person,
        prov.Agent,
        prov.Organization,
        ResourceType,
        BlankNodeType,
        List[
            Union[
                foaf.Agent,
                foaf.Organization,
                foaf.Person,
                prov.Person,
                prov.Agent,
                prov.Organization,
                ResourceType,
                BlankNodeType
            ]
        ]
    ] = None  # dcterms:creator
    publisher: Union[foaf.Agent, List[foaf.Agent]] = None  # dcterms:publisher
    contributor: Union[foaf.Agent, List[foaf.Agent]] = None  # dcterms:contributor
    license: Optional[Union[ResourceType, List[ResourceType]]] = None  # dcat:license
    version: str = None  # dcat:version
    identifier: str = None  # dcterms:identifier
    hasPart: Optional[Union[ResourceType, List[ResourceType]]] = Field(default=None, alias='has_part')
    keyword: Optional[Union[str, List[str]]] = None  # dcat:keyword

    @model_validator(mode="before")
    def change_id(self):
        """Change the id to the downloadURL"""
        return as_id(self, "identifier")

    @field_validator('identifier', mode='before')
    @classmethod
    def _identifier(cls, identifier):
        """parse datetime"""
        if identifier.startswith('http'):
            return str(HttpUrl(identifier))
        return identifier


@namespaces(dcat="http://www.w3.org/ns/dcat#")
@urirefs(DataService='dcat:DataService',
         endpointURL='dcat:endpointURL',
         servesDataset='dcat:servesDataset')
class DataService(Resource):
    endpointURL: Union[HttpUrl, FileUrl] = Field(alias='endpoint_url', default=None)  # dcat:endpointURL
    servesDataset: "Dataset" = Field(alias='serves_dataset', default=None)  # dcat:servesDataset



@namespaces(dcat="http://www.w3.org/ns/dcat#",
            prov="http://www.w3.org/ns/prov#",
            dcterms="http://purl.org/dc/terms/")
@urirefs(Distribution='dcat:Distribution',
         downloadURL='dcat:downloadURL',
         accessURL='dcat:accessURL',
         mediaType='dcat:mediaType',
         byteSize='dcat:byteSize',
         hasPart='dcterms:hasPart',
         checksum='spdx:checksum',
         accessService='dcat:distribution'
         )
class Distribution(Resource):
    """Implementation of dcat:Distribution

    .. note::
        More than the below parameters are possible but not explicitly defined here.


    Parameters
    ----------
    downloadURL: Union[HttpUrl, FileUrl]
        Download URL of the distribution (dcat:downloadURL)
    mediaType: HttpUrl = None
        Media type of the distribution (dcat:mediaType).
        Should be defined by the [IANA Media Types registry](https://www.iana.org/assignments/media-types/media-types.xhtml)
    byteSize: int = None
        Size of the distribution in bytes (dcat:byteSize)
    """
    downloadURL: Union[HttpUrl, FileUrl, pathlib.Path] = Field(default=None, alias='download_URL')
    accessURL: Union[HttpUrl, FileUrl, pathlib.Path] = Field(default=None, alias='access_URL')
    mediaType: Union[str, ResourceType] = Field(default=None, alias='media_type')  # dcat:mediaType
    byteSize: int = Field(default=None, alias='byte_size')  # dcat:byteSize
    hasPart: Union[Thing, List[Thing]] = Field(default=None, alias='has_part')  # dcterms:hasPart
    checksum: Union[ResourceType, Checksum] = None  # spdx:checksum
    accessService: DataService = Field(default=None, alias='access_service')  # dcat:accessService

    def _repr_html_(self):
        """Returns the HTML representation of the class"""
        if self.download_URL is not None:
            return f"{self.__class__.__name__}({self.download_URL})"
        return super()._repr_html_()

    def download(self,
                 dest_filename: Union[str, pathlib.Path] = None,
                 overwrite_existing: bool = False,
                 **kwargs) -> pathlib.Path:
        """Downloads the distribution
        kwargs are passed to the download_file function, which goes to requests.get()"""
        if dest_filename is not None:
            if pathlib.Path(dest_filename).is_dir():
                raise IsADirectoryError(f'Destination filename {dest_filename} is a directory')
        if "exist_ok" in kwargs:
            overwrite_existing = kwargs.pop("exist_ok")

        if self.download_URL is None:
            raise ValueError('The downloadURL is not defined')

        downloadURL = str(self.downloadURL)
        def _parse_file_url(furl):
            """in windows, we might need to strip the leading slash"""
            fname = pathlib.Path(furl)
            if fname.exists():
                return fname
            fname = pathlib.Path(self.download_URL.path[1:])
            if fname.exists():
                return fname
            raise FileNotFoundError(f'File {self.download_URL.path} does not exist')

        if self.download_URL.scheme == 'file':
            if dest_filename is None:
                return _parse_file_url(self.download_URL.path)
            else:
                return shutil.copy(_parse_file_url(self.download_URL.path), dest_filename)
        if dest_filename is None:
            import os
            from urllib.parse import urlparse
            dest_filename = os.path.basename(urlparse(downloadURL).path)
            if dest_filename == '':
                dest_filename = downloadURL.rsplit("#", 1)[-1]

        dest_filename = pathlib.Path(dest_filename)
        if not dest_filename.suffix.startswith("."):
            raise ValueError('Destination filename must have a valid suffix/extension')

        if dest_filename.exists():
            return dest_filename
        return download_file(self.download_URL,
                             dest_filename,
                             exist_ok=overwrite_existing,
                             **kwargs)

    @field_validator('mediaType', mode='before')
    @classmethod
    def _mediaType(cls, mediaType):
        """should be a valid URI, like: https://www.iana.org/assignments/media-types/text/markdown"""
        if isinstance(mediaType, str):
            if mediaType.startswith('http'):
                return mediaType
            elif mediaType.startswith('iana:'):
                return "https://www.iana.org/assignments/media-types/" + mediaType.split(":", 1)[-1]
            elif re.match('[a-z].*/[a-z].*', mediaType):
                return "https://www.iana.org/assignments/media-types/" + mediaType
        return mediaType

    @field_validator('downloadURL', mode='before')
    @classmethod
    def _downloadURL(cls, downloadURL):
        """a pathlib.Path is also allowed but needs to be converted to a URL"""
        if isinstance(downloadURL, pathlib.Path):
            return FileUrl(f'file://{downloadURL.resolve().absolute()}')
        return downloadURL


@namespaces(dcat="http://www.w3.org/ns/dcat#")
@urirefs(DatasetSeries='dcat:DatasetSeries')
class DatasetSeries(Resource):
    """Pydantic implementation of dcat:DatasetSeries"""


@namespaces(dcat="http://www.w3.org/ns/dcat#",
            prov="http://www.w3.org/ns/prov#",
            dcterms="http://purl.org/dc/terms/")
@urirefs(Dataset='dcat:Dataset',
         identifier='dcterms:identifier',
         creator='dcterms:creator',
         distribution='dcat:distribution',
         modified='dcterms:modified',
         landingPage='dcat:landingPage',
         inSeries='dcat:inSeries',
         license='dcterms:license', )
class Dataset(Resource):
    """Pydantic implementation of dcat:Dataset

    .. note::

        More than the below parameters are possible but not explicitly defined here.



    Parameters
    ----------
    title: str
        Title of the resource (dcterms:title)
    description: str = None
        Description of the resource (dcterms:description)
    version: str = None
        Version of the resource (dcat:version)
    identifier: HttpUrl = None
        Identifier of the resource (dcterms:identifier)
    distribution: List[Distribution] = None
        Distribution of the resource (dcat:Distribution)
    landingPage: HttpUrl = None
        Landing page of the resource (dcat:landingPage)
    modified: datetime = None
        Last modified date of the resource (dcterms:modified)
    inSeries: DatasetSeries = None
        The series the dataset belongs to (dcat:inSeries)
    """
    identifier: ResourceType = None  # dcterms:identifier, see https://www.w3.org/TR/vocab-dcat-3/#ex-identifier
    # http://www.w3.org/ns/prov#Person, see https://www.w3.org/TR/vocab-dcat-3/#ex-adms-identifier
    distribution: Union[Distribution, List[Distribution]] = None  # dcat:Distribution
    modified: datetime = None  # dcterms:modified
    landingPage: HttpUrl = Field(default=None, alias='landing_page')  # dcat:landingPage
    inSeries: DatasetSeries = Field(default=None, alias='in_series')  # dcat:inSeries
    license: Optional[Union[ResourceType, List[ResourceType]]] = None  # dcat:license

    @field_validator('modified', mode='before')
    @classmethod
    def _modified(cls, modified):
        """parse datetime"""
        if isinstance(modified, str):
            return parser.parse(modified)
        return modified


DataService.model_rebuild()


from dataclasses import dataclass

from pydantic import HttpUrl


@dataclass
class URIValue:
    """A class to represent a value with an associated URI."""
    value: str  # the value of the URI
    namespace: str  # the URI that identifies the predicate of the value
    prefix: str  # TODO: this might be None and figured out automatically based on known namespaces, for now it's a required field

    def __post_init__(self):
        # verify that the URI is a valid URL
        try:
            HttpUrl(self.namespace)
        except ValueError as e:
            raise ValueError(f"Invalid namespace URI: {self.namespace}. Error: {e}")


if __name__ == "__main__":
    # Example usage
    uri_value = URIValue(value="Example Value", namespace="http://example.com/resource", prefix="ex")
    print(uri_value)



import logging
import unittest
from typing import List
from typing import Union

import rdflib
from pydantic import Field, HttpUrl

from ontolutils import Property
from ontolutils import Thing, urirefs, namespaces, build
from ontolutils.namespacelib import HDF5
LOG_LEVEL = logging.DEBUG


class TestHDF5(unittest.TestCase):

    def test_dataset(self):
        @namespaces(schema="https://schema.org/",
                    hdf=str(HDF5))
        @urirefs(Dataset='hdf:Dataset',
                 name='hdf:name',
                 datatype='hdf:datatype')
        class Dataset(Thing):
            name: str = Field(..., description="Name of the dataset")
            datatype: HttpUrl = Field(..., description="Datatype of the dataset")

        ds = Dataset(name="test", datatype=HDF5.H5T_INTEGER)
        serialization = ds.serialize(format="ttl")
        expectation = """@prefix hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#> .

[] a hdf:Dataset ;
    hdf:datatype hdf:H5T_INTEGER ;
    hdf:name "test" .

"""
        self.assertEqual(serialization, expectation)

        # ds = Dataset(name="test", datatype=HDF5.H5T_INTEGER)
        ds = Dataset(name="test", datatype="https://example/integer")
        serialization = ds.serialize(format="ttl")
        expectation = """@prefix hdf: <http://purl.allotrope.org/ontologies/hdf5/1.8#> .

[] a hdf:Dataset ;
    hdf:datatype <https://example/integer> ;
    hdf:name "test" .

"""
        self.assertEqual(serialization, expectation)



import json
import logging
import pathlib
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, Union, Optional, Any, List, Type
from urllib.parse import urlparse

import rdflib
import yaml
from pydantic import AnyUrl, HttpUrl, FileUrl, BaseModel, Field, field_validator, model_validator
from pydantic import field_serializer
from pydantic_core import Url
from rdflib import XSD
from rdflib.plugins.shared.jsonld.context import Context

from .decorator import urirefs, namespaces, URIRefManager, NamespaceManager, _is_http_url
from .thingmodel import ThingModel
from .utils import split_URIRef
from .. import get_config
from ..typing import BlankNodeType

logger = logging.getLogger('ontolutils')
URL_SCHEMES = {"http", "https", "urn", "doi"}


@dataclass
class Property:
    name: str
    property_type: Any
    default: Optional[Any] = None
    namespace: Optional[HttpUrl] = None
    namespace_prefix: Optional[str] = None

    def __post_init__(self):
        if self.namespace is None and self.namespace_prefix is not None:
            raise ValueError("If namespace_prefix is given, then namespace must be given as well.")
        if self.namespace_prefix is None and self.namespace is not None:
            raise ValueError("If namespace is given, then namespace_prefix must be given as well.")


def resolve_iri(key_or_iri: str, context: Context) -> Optional[str]:
    """Resolve a key or IRI to a full IRI using the context."""
    if key_or_iri.startswith('http'):
        return str(key_or_iri)
    if ':' in key_or_iri:
        iri = context.resolve(key_or_iri)
        if iri.startswith('http'):
            return iri
    try:
        return context.terms.get(key_or_iri).id
    except AttributeError:
        if key_or_iri == 'label':
            return 'http://www.w3.org/2000/01/rdf-schema#label'
    return


def _get_n3():
    return rdflib.BNode().n3()


def build_blank_n3() -> str:
    return rdflib.BNode().n3()


def build_blank_id() -> str:
    id_generator = get_config("blank_id_generator")
    if id_generator is None:
        id_generator = _get_n3

    _blank_node = id_generator()
    return _blank_node


def is_url(iri: str) -> bool:
    try:
        s = str(iri)
        scheme = urlparse(s).scheme.lower()
        if scheme in URL_SCHEMES:
            try:
                AnyUrl(iri)
                return True
            except Exception:
                return False
        return False
    except Exception:
        return False


# class LangString(BaseModel):
#     value: Union[str, int, float]
#     lang: Optional[str] = None
#     datatype: Optional[Union[HttpUrl, str]] = None
#
#     # Validate the datatype itself
#     @field_validator('datatype', mode='before')
#     @classmethod
#     def validate_datatype(cls, datatype):
#         if datatype is None:
#             return datatype
#         # accept either HttpUrl objects or strings that parse as HttpUrl
#         if not is_url(datatype):
#             raise ValueError(f"The datatype must be a valid IRI but got {datatype}.")
#         return datatype
#
#     # Enforce: lang XOR datatype (not both set)
#     @model_validator(mode='after')
#     def check_lang_xor_datatype(self):
#         if self.lang and self.datatype:
#             raise ValueError("A LangString cannot have both a datatype and a language.")
#         return self
#
#     def __str__(self) -> str:
#         return str(self.value)

# A light, permissive BCP-47-ish check: en | en-US | zh-Hant | de-CH-1996, etc.
def _looks_like_lang(tag: str) -> bool:
    import re
    return bool(re.fullmatch(r"[A-Za-z]{1,8}(?:-[A-Za-z0-9]{1,8})*", tag))


def _split_value_lang(s: str) -> tuple[str, Optional[str]]:
    """
    Split 'value@lang' only if the suffix looks like a language tag and there is no trailing space.
    Otherwise, return (s, None).
    """
    if "@" not in s:
        return s, None
    # Take the last '@' so values with earlier '@' (e.g., emails) aren't split incorrectly
    head, tail = s.rsplit("@", 1)
    if not head:  # avoid '@en' case
        return s, None
    if " " in tail:  # lang tags shouldn't contain spaces
        return s, None
    if _looks_like_lang(tail):
        return head, tail
    return s, None


class LangString(BaseModel):
    """Language-String"""
    value: str
    lang: Optional[str] = None

    # Accept str, dict, rdflib.Literal, or LangString
    @model_validator(mode="before")
    @classmethod
    def coerce_input(cls, v: Any):

        if isinstance(v, cls):
            return v

        if isinstance(v, rdflib.Literal):
            return {"value": str(v), "lang": v.language}

        if isinstance(v, dict):
            return v

        if isinstance(v, str):
            value, lang = _split_value_lang(v)
            return {"value": value, "lang": lang}

        if isinstance(v, list):
            return [cls.model_validate(_v) for _v in v]

        return v

    def __hash__(self):
        return hash((self.value, self.lang))

    def __str__(self, show_lang: bool = None):
        if show_lang is None:
            show_lang = get_config("show_lang_in_str")
        if self.lang and show_lang:
            return f"{self.value}@{self.lang}"
        return f"{self.value}" if self.lang else str(self.value)

    def to_dict(self):
        return {"value": self.value, "lang": self.lang}

    @field_serializer("value", "lang")
    def _identity(self, v):
        return v

    def __eq__(self, other):
        """Equality comparison with another LangString or a plain string.

        Examples of equality:
        >>> LangString(value="Hello", lang="en") == LangString(value="Hello", lang="en")
        True
        >>> LangString(value="Hello", lang="en") == "Hello@en"
        True
        >>> LangString(value="Hello", lang="en") == "Hello"
        True
        >>> LangString(value="Hello") == "Hello"
        True
        >>> LangString(value="Hello") == LangString(value="Hello")
        True
        >>> LangString(value="Hello", lang="en") == "Hello@fr"
        False
        """
        if isinstance(other, LangString):
            return self.value == other.value and self.lang == other.lang
        if isinstance(other, str):
            return str(self) == other or self.value == other
        raise TypeError(f"Cannot compare LangString with {type(other)}")


def langstring_representer(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


yaml.add_representer(LangString, langstring_representer)


def serialize_lang_str_field(lang_str: LangString):
    if lang_str.lang is None:
        return lang_str.value
    result = {"@value": lang_str.value, "@language": lang_str.lang}
    return result


def datetime_to_literal(dt: datetime):
    """Turn a datetime into a literal."""
    return {
        "@value": dt.isoformat(),
        "@type": str(XSD.dateTime if dt.hour or dt.minute or dt.second else XSD.date)
    }


def _parse_string_value(value, ctx):
    if is_url(value):
        return {"@id": str(value)}
    elif ":" in value:
        _ns, _value = value.split(":", 1)
        if _ns in ctx:
            return {"@id": f"{ctx[_ns]}{_value}"}
    return value


@namespaces(owl='http://www.w3.org/2002/07/owl#',
            rdfs='http://www.w3.org/2000/01/rdf-schema#',
            dcterms='http://purl.org/dc/terms/',
            schema='https://schema.org/',
            skos='http://www.w3.org/2004/02/skos/core#',
            )
@urirefs(Thing='owl:Thing',
         label='rdfs:label',
         about='schema:about',
         relation='dcterms:relation',
         closeMatch='skos:closeMatch',
         exactMatch='skos:exactMatch')
class Thing(ThingModel):
    """Most basic concept class owl:Thing (see also https://www.w3.org/TR/owl-guide/)

    This class is basis to model other concepts.

    Example for `prov:Person`:

    >>> @namespaces(prov='https://www.w3.org/ns/prov#',
    >>>             foaf='http://xmlns.com/foaf/0.1/')
    >>> @urirefs(Person='prov:Person', first_name='foaf:firstName')
    >>> class Person(Thing):
    >>>     first_name: str = None
    >>>     last_name: str = None

    >>> p = Person(first_name='John', last_name='Doe', age=30)
    >>> # Note, that age is not defined in the class! This is allowed, but may not be
    >>> # serialized into an IRI although the ontology defines it

    >>> print(p.model_dump_jsonld())
    >>> {
    >>>     "@context": {
    >>>         "prov": "https://www.w3.org/ns/prov#",
    >>>         "foaf": "http://xmlns.com/foaf/0.1/",
    >>>         "first_name": "foaf:firstName"
    >>>     },
    >>>     "@id": "N23036f1a4eb149edb7db41b2f5f4268c",
    >>>     "@type": "prov:Person",
    >>>     "foaf:firstName": "John",
    >>>     "last_name": "Doe",
    >>>     "age": "30"  # Age appears as a field without context!
    >>> }

    Note, that values are validated, as `Thing` is a subclass of `pydantic.BaseModel`:

    >>> Person(first_name=1)

    Will lead to a validation error:

    >>> # Traceback (most recent call last):
    >>> # ...
    >>> # pydantic_core._pydantic_core.ValidationError: 1 validation error for Person
    >>> # first_name
    >>> #   Input should be a valid string [type=string_type, input_value=1, input_type=int]
    >>> #     For further information visit https://errors.pydantic.dev/2.4/v/string_type

    """
    id: Optional[Union[str, HttpUrl, FileUrl, BlankNodeType, None]] = Field(default_factory=build_blank_id)  # @id
    label: Optional[Union[LangString, List[LangString]]] = None  # rdfs:label
    about: Optional[
        Union[
            str, HttpUrl, FileUrl, ThingModel, BlankNodeType, List[Union[HttpUrl, FileUrl, ThingModel, BlankNodeType]]]
    ] = None  # schema:about
    relation: Optional[Union[HttpUrl, FileUrl, BlankNodeType, ThingModel]] = None
    closeMatch: Optional[Union[HttpUrl, FileUrl, BlankNodeType, ThingModel]] = None
    exactMatch: Optional[Union[HttpUrl, FileUrl, BlankNodeType, ThingModel]] = None

    # class Config:
    #     arbitrary_types_allowed = True

    @property
    def namespace(self) -> str:
        compact_uri = self.urirefs[self.__class__.__name__]
        prefix, name = compact_uri.split(':')
        return self.namespaces[prefix]

    @property
    def uri(self) -> str:
        compact_uri = self.urirefs[self.__class__.__name__]
        prefix, name = compact_uri.split(':')
        namespace = self.namespaces[prefix]
        return f"{namespace}{name}"

    def map(self, other: Type[ThingModel]) -> ThingModel:
        """Return the class as another class. This is useful to convert a ThingModel
        to another ThingModel class."""
        if not issubclass(other, ThingModel):
            raise TypeError(f"Cannot convert {self.__class__} to {other}. "
                            f"{other} must be a subclass of ThingModel.")
        combined_urirefs = {**self.urirefs, **URIRefManager[other]}
        combined_urirefs.pop(self.__class__.__name__)
        URIRefManager.data[other] = combined_urirefs

        combined_namespaces = {**self.namespaces, **NamespaceManager[other]}
        NamespaceManager.data[other] = combined_namespaces
        return other(**self.model_dump(exclude_none=True))

    @field_validator('id', mode="before")
    @classmethod
    def _id(cls, id: Optional[Union[str, HttpUrl, FileUrl, BlankNodeType]]) -> str:
        if id is None:
            return build_blank_n3()
        if isinstance(id, rdflib.URIRef):
            return id.n3()
        if isinstance(id, rdflib.BNode):
            return id.n3()
        if not isinstance(id, str):
            raise TypeError(
                f'The ID must be a string, HttpUrl, FileUrl or BlankNodeType but got {type(id)}.'
            )
        if id.startswith('http'):
            return str(HttpUrl(id))
        if id.startswith('file'):
            return str(FileUrl(id))
        if id.startswith('_:'):
            return id  # this is a blank node
        raise ValueError(
            f'The ID must be a valid IRI or blank node but got "{id}". '
            'It must start with "_:", "http", "file".'
        )

    @classmethod
    def build(cls, namespace: HttpUrl,
              namespace_prefix: str,
              class_name: str,
              properties: List[Union[Property, Dict]]) -> type:
        """Build a Thing object"""
        return build(
            namespace,
            namespace_prefix,
            class_name,
            properties,
            cls
        )

    def __lt__(self, other: ThingModel) -> bool:
        """Less than comparison. Useful to sort Thing objects.
        Comparison can only be done with other Thing objects and if an ID is given.
        If one of the objects has no ID, then False is returned."""
        if not isinstance(other, ThingModel):
            raise TypeError(f"Cannot compare {self.__class__} with {type(other)}")
        if self.id is None or other.id is None:
            return False
        return self.id <= other.id

    def get_jsonld_dict(self,
                        base_uri: Optional[Union[str, AnyUrl]] = None,
                        context: Optional[Union[Dict, str]] = None,
                        exclude_none: bool = True,
                        resolve_keys: bool = False,
                        ) -> Dict:
        """Return the JSON-LD dictionary of the object. This will include the context
        and the fields of the object.

        Parameters
        ----------
        context: Optional[Union[Dict, str]]
            The context to use for the JSON-LD serialization. If a string is given, it will
            be interpreted as an import statement and will be added to the context.
        exclude_none: bool=True
            Exclude fields with None values
        resolve_keys: bool=False
            If True, then attributes of a Thing class will be resolved to the full IRI and
            explained in the context.
        base_uri: Optional[Union[str, AnyUrl]]=None
            The base URI to use for blank nodes (only used if no ID is set).
            This is useful, because blank nodes are not globally unique and
            can lead to problems when merging data from different sources.

            Example:

                In the following example, first_name refers to foaf:firstName:

                >>> @namespaces(foaf='http://xmlns.com/foaf/0.1/')
                >>> @urirefs(Person='foaf:Person', first_name='foaf:firstName')
                >>> class Person(Thing):
                >>>     first_name: str = None

                >>> p = Person(first_name='John')
                >>> p.model_dump_jsonld(resolve_keys=True)

                This will result "first_name": "foaf:firstName" showing up in the context:

                >>> {
                >>>     "@context": {
                >>>         "foaf": "http://xmlns.com/foaf/0.1/",
                >>>         "first_name": "foaf:firstName"
                >>>     },
                >>>     "@type": "foaf:Person",
                >>>     "foaf:firstName": "John"
                >>> }

                While resolve_keys=False will result in:

                >>> {
                >>>     "@context": {
                >>>         "foaf": "http://xmlns.com/foaf/0.1/"
                >>>     },
                >>>     "@type": "foaf:Person",
                >>>     "foaf:firstName": "John"
                >>> }


        Returns
        -------
        Dict
            The JSON-LD dictionary
        """
        from .urivalue import URIValue
        logger.debug('Initializing RDF graph to dump the Thing to JSON-LD')

        # lets auto-generate the context
        at_context: Dict = NamespaceManager.get(self.__class__, {}).copy()

        if context is None:
            context = {}

        if not isinstance(context, dict):
            raise TypeError(f"Context must be a dict, not {type(context)}")

        at_context.update(**context)

        # ctx = Context(source={**at_context, **URIRefManager.get(self.__class__)})

        logger.debug(f'The context is "{at_context}".')

        def _serialize_fields(
                obj: Union[ThingModel, int, str, float, bool, datetime],
                _exclude_none: bool
        ) -> Union[Dict, int, str, float, bool]:
            """Serializes the fields of a Thing object into a json-ld
            dictionary (without context!). Note, that IDs can automatically be
            generated (with a local prefix)

            Parameter
            ---------
            obj: Union[ThingModel, int, str, float, bool, datetime]
                The object to serialize (a subclass of ThingModel). All other types will
                be returned as is. One exception is datetime, which will be serialized
                to an ISO string.
            _exclude_none: bool=True
                If True, fields with None values will be excluded from the
                serialization

            Returns
            -------
            Union[Dict, int, str, float, bool]
                The serialized fields or the object as is
            """

            obj_ctx = Context(source={**context,
                                      **NamespaceManager.get(obj.__class__, {}),
                                      **URIRefManager.get(obj.__class__, {})})

            if isinstance(obj, str) and _is_http_url(obj):
                return {"@id": str(obj)}
            if isinstance(obj, str):
                return _parse_string_value(obj, at_context)
            if isinstance(obj, Url):
                return {"@id": str(obj)}
            if isinstance(obj, list):
                return [_serialize_fields(o, _exclude_none) for o in obj]
            if isinstance(obj, (int, float, bool)):
                return obj
            if isinstance(obj, LangString):
                return serialize_lang_str_field(obj)
            if isinstance(obj, datetime):
                return datetime_to_literal(obj)

            uri_ref_manager = URIRefManager.get(obj.__class__, None)
            at_context.update(NamespaceManager.get(obj.__class__, {}))

            if isinstance(obj, ThingModel):
                if obj.model_extra:
                    for extra in obj.model_extra.values():
                        if isinstance(extra, URIValue):
                            at_context[extra.prefix] = extra.namespace

            if uri_ref_manager is None:
                return str(obj)

            try:
                serialized_fields = {}
                if isinstance(obj, ThingModel):
                    if obj.model_extra:
                        for extra_field_name, extra_field_value in obj.model_extra.items():
                            if isinstance(extra_field_value, URIValue):
                                serialized_fields[extra_field_name] = f"{extra_field_value.prefix}:{extra_field_name}"
                for k in obj.model_dump(exclude_none=_exclude_none):
                    value = getattr(obj, k)
                    if isinstance(value, str):
                        value = _replace_context_url_with_prefix(value, at_context)
                    if value is not None and k not in ('id', '@id'):
                        iri = uri_ref_manager.get(k, k)
                        if _is_http_url(iri):
                            serialized_fields[iri] = value
                        if resolve_keys:
                            serialized_fields[iri] = value
                        else:
                            term = obj_ctx.find_term(obj_ctx.expand(iri))
                            if term:
                                if obj_ctx.shrink_iri(term.id).split(':')[1] != k:
                                    at_context[k] = term.id
                                    serialized_fields[k] = value
                                else:
                                    serialized_fields[iri] = value
            except AttributeError as e:
                raise AttributeError(f"Could not serialize {obj} ({obj.__class__}). Orig. err: {e}") from e

            # datetime
            for k, v in serialized_fields.copy().items():
                _field = serialized_fields.pop(k)
                key = k
                if isinstance(v, Thing):
                    serialized_fields[key] = _serialize_fields(v, _exclude_none=_exclude_none)
                elif isinstance(v, list):
                    serialized_fields[key] = [
                        _serialize_fields(i, _exclude_none=_exclude_none) for i in v]
                elif isinstance(v, (int, float)):
                    serialized_fields[key] = v
                elif isinstance(v, LangString):
                    serialized_fields[key] = serialize_lang_str_field(v)
                elif _is_http_url(v):
                    serialized_fields[key] = {"@id": str(v)}
                elif isinstance(v, URIValue):
                    serialized_fields[f"{v.prefix}:{key}"] = v.value
                elif isinstance(v, datetime):
                    serialized_fields[key] = datetime_to_literal(v)
                elif isinstance(v, str):
                    serialized_fields[key] = _parse_string_value(v, at_context)
                else:
                    serialized_fields[key] = _serialize_fields(v, _exclude_none=_exclude_none)

            _type = URIRefManager[obj.__class__].get(obj.__class__.__name__, obj.__class__.__name__)

            out = {"@type": _type, **serialized_fields}
            # if no ID is given, generate a local one:
            if obj.id is not None:
                out["@id"] = _replace_context_url_with_prefix(_parse_blank_node(obj.id, base_uri), context)
            else:
                out["@id"] = _replace_context_url_with_prefix(_parse_blank_node(rdflib.BNode().n3(), base_uri), context)
            return out

        serialization = _serialize_fields(self, _exclude_none=exclude_none)

        jsonld = {
            "@context": at_context,
            **serialization
        }

        properties = self.__class__.model_json_schema().get("properties", {})
        if not properties:
            properties = self.__class__.model_json_schema().get("items", {}).get(self.__class__.__name__, {}).get(
                "properties", {})

        for field_name, field_value in properties.items():
            _use_as_id = field_value.get("use_as_id", None)
            if _use_as_id is not None:
                warnings.warn("The use_as_id field is deprecated. Use the @id field instead.", DeprecationWarning)
                _id = getattr(self, field_name)
                if _id is not None:
                    if str(_id).startswith(("_:", "http")):
                        jsonld["@id"] = getattr(self, field_name)
                    else:
                        raise ValueError(f'The ID must be a valid IRI or blank node but got "{_id}".')
        return jsonld

    def serialize(self,
                  format: str,
                  context: Optional[Dict] = None,
                  exclude_none: bool = True,
                  resolve_keys: bool = True,
                  base_uri: Optional[Union[str, AnyUrl]] = None,
                  **kwargs) -> str:
        """
        Serialize the object to a given format. This method calls rdflib.Graph().parse(),
        so the available formats are the same as for the rdflib library:
            ``"xml"``, ``"n3"``,
           ``"turtle"``, ``"nt"``, ``"pretty-xml"``, ``"trix"``, ``"trig"``,
           ``"nquads"``, ``"json-ld"`` and ``"hext"`` are built in.
        The kwargs are passed to rdflib.Graph().parse()
        """

        jsonld_dict = self.get_jsonld_dict(
            context=context,
            exclude_none=exclude_none,
            resolve_keys=resolve_keys,
            base_uri=base_uri
        )
        jsonld_str = json.dumps(jsonld_dict)

        logger.debug(f'Parsing the following jsonld dict to the RDF graph: {jsonld_str}')
        g = rdflib.Graph()

        if context:
            for k, v in context.items():
                g.bind(k, rdflib.Namespace(v))

        g.parse(data=jsonld_str, format='json-ld')

        _context = jsonld_dict.get('@context', {})
        if context:
            _context.update(context)

        return g.serialize(format=format,
                           context=_context,
                           **kwargs)

    def model_dump_jsonld(
            self,
            context: Optional[Dict] = None,
            exclude_none: bool = True,
            rdflib_serialize: bool = False,
            resolve_keys: bool = True,
            base_uri: Optional[Union[str, AnyUrl]] = None,
            indent: int = 4) -> str:
        """Similar to model_dump_json() but will return a JSON string with
        context resulting in a JSON-LD serialization. Using `rdflib_serialize=True`
        will use the rdflib to serialize. This will make the output a bit cleaner
        but is not needed in most cases and just takes a bit more time (and requires
        an internet connection.

        Note, that if `rdflib_serialize=True`, then a blank node will be generated if no ID is set.

        Parameters
        ----------
        context: Optional[Union[Dict, str]]
            The context to use for the JSON-LD serialization. If a string is given, it will
            be interpreted as an import statement and will be added to the context.
        exclude_none: bool=True
            Exclude fields with None values
        rdflib_serialize: bool=False
            If True, the output will be serialized using rdflib. This results in a cleaner
            output but is not needed in most cases and just takes a bit more time (and requires
            an internet connection). Will also generate a blank node if no ID is set.
        resolve_keys: bool=False
            If True, then attributes of a Thing class will be resolved to the full IRI and
            explained in the context.
        indent: int=4
            The indent of the JSON-LD string
        base_uri: Optional[HttpUrl]=None
            The base URI to use for blank nodes (only used if no ID is set).
            This is useful, because blank nodes are not globally unique and
            can lead to problems when merging data from different sources.

            .. seealso:: `Thing.get_jsonld_dict`

        Returns
        -------
        str
            The JSON-LD string
        """
        jsonld_dict = self.get_jsonld_dict(
            context=context,
            exclude_none=exclude_none,
            resolve_keys=resolve_keys,
            base_uri=base_uri
        )
        jsonld_str = json.dumps(jsonld_dict, indent=4)
        if not rdflib_serialize:
            return jsonld_str

        logger.debug(f'Parsing the following jsonld dict to the RDF graph: {jsonld_str}')
        g = rdflib.Graph()
        g.parse(data=jsonld_str, format='json-ld')

        _context = jsonld_dict.get('@context', {})
        if context:
            _context.update(context)

        return g.serialize(format='json-ld',
                           context=_context,
                           indent=indent)

    def model_dump_ttl(self,
                       context: Optional[Dict] = None,
                       exclude_none: bool = True,
                       resolve_keys: bool = True,
                       base_uri: Optional[Union[str, AnyUrl]] = None
                       ):
        """Dump the model as a Turtle string."""
        return self.serialize(
            format="turtle",
            context=context,
            exclude_none=exclude_none,
            resolve_keys=resolve_keys,
            base_uri=base_uri
        )

    def __repr__(self, limit: Optional[int] = None):
        _fields = {k: getattr(self, k) for k in self.__class__.model_fields.keys() if getattr(self, k) is not None}
        if self.model_extra:
            repr_extra = ", ".join([f"{k}={v}" for k, v in {**_fields, **self.model_extra}.items()])
        else:
            repr_extra = ", ".join([f"{k}={v}" for k, v in {**_fields}.items()])
        if limit is None or len(repr_extra) < limit:
            return f"{self.__class__.__name__}({repr_extra})"
        return f"{self.__class__.__name__}({repr_extra[0:limit]}...)"

    def __str__(self, limit: Optional[int] = None):
        return self.__repr__(limit=limit)

    def _repr_html_(self) -> str:
        """Returns the HTML representation of the class"""
        # _fields = {k: getattr(self, k) for k in self.model_fields if getattr(self, k) is not None}
        # repr_fields = ", ".join([f"{k}={v}" for k, v in _fields.items()])
        return self.__repr__()

    @classmethod
    def from_jsonld(cls,
                    source: Optional[Union[str, pathlib.Path]] = None,
                    data: Optional[Union[str, Dict]] = None,
                    limit: Optional[int] = None,
                    context: Optional[Dict] = None):
        """Initialize the class from a JSON-LD source

        Note the inconsistency in the schema.org protocol. Codemeta for instance uses http whereas
        https is the current standard. This repo only works with https. If you have a http schema,
        this method will replace http with https.

        Parameters
        ----------
        source: Optional[Union[str, pathlib.Path]]=None
            The source of the JSON-LD data (filename). Must be given if data is None.
        data: Optional[Union[str, Dict]]=None
            The JSON-LD data as a str or dictionary. Must be given if source is None.
        limit: Optional[int]=None
            The limit of the number of objects to return. If None, all objects will be returned.
            If limit is 1, then the first object will be returned, else a list of objects.
        context: Optional[Dict]=None
            The context to use for the JSON-LD serialization. If a string is given, it will
            be interpreted as an import statement and will be added to the context.

        """
        from . import query
        if data is not None:
            if isinstance(data, dict):
                data = json.dumps(data)
            if 'http://schema.org/' in data:
                warnings.warn('Replacing http with https in the JSON-LD data. '
                              'This is a workaround for the schema.org inconsistency.',
                              UserWarning)
                data = data.replace('http://schema.org/', 'https://schema.org/')
        return query(cls, source=source, data=data, limit=limit, context=context)

    @classmethod
    def iri(cls, key: str = None, compact: bool = False):
        """Return the IRI of the class or the key

        Parameter
        ---------
        key: str
            The key (field) of the class
        compact: bool
            If True, returns the short form of the IRI, e.g. 'owl:Thing'
            If False, returns the full IRI, e.g. 'http://www.w3.org/2002/07/owl#Thing'

        Returns
        -------
        str
            The IRI of the class or the key, e.g. 'http://www.w3.org/2002/07/owl#Thing' or
            'owl:Thing' if compact is True
        """
        if key is None:
            iri_short = URIRefManager[cls][cls.__name__]
        else:
            iri_short = URIRefManager[cls][key]
        if compact:
            return iri_short
        ns, key = split_URIRef(iri_short)
        ns_iri = NamespaceManager[cls].get(ns, None)
        return f'{ns_iri}{key}'

    @property
    def namespaces(self) -> Dict:
        """Return the namespaces of the class"""
        return get_namespaces(self.__class__)

    @property
    def urirefs(self) -> Dict:
        """Return the urirefs of the class"""
        return get_urirefs(self.__class__)

    @classmethod
    def get_context(cls) -> Dict:
        """Return the context of the class"""
        return get_namespaces(cls)


def _replace_context_url_with_prefix(value: str, context: Dict) -> str:
    for context_key, context_url in context.items():
        if value.startswith(context_url):
            return value.replace(context_url, context_key + ':')
    return value


def _parse_blank_node(_id, base_uri: Optional[Union[str, AnyUrl]]):
    if base_uri:
        base_uri = AnyUrl(base_uri)
    if base_uri is None:
        return _id
    if isinstance(_id, rdflib.BNode):
        return f"{base_uri}{_id}"
    if isinstance(_id, str) and _id.startswith('_:'):
        return f"{base_uri}{_id[2:]}"
    if isinstance(_id, str) and _id.startswith('http'):
        return _id
    if isinstance(_id, str) and _id.startswith('file://'):
        return _id
    if isinstance(_id, str) and _id.startswith('urn:'):
        return _id
    if isinstance(_id, str) and _id.startswith('bnode:'):
        return f"{base_uri}{_id[6:]}"
    if isinstance(_id, str) and _id.startswith('N'):
        # This is a blank node generated by rdflib
        return f"{base_uri}{_id}"
    warnings.warn(f"Could not parse blank node ID '{_id}'. ")
    return _id


def get_urirefs(cls: Thing) -> Dict:
    """Return the URIRefs of the class"""
    return URIRefManager[cls]


def get_namespaces(cls: Thing) -> Dict:
    """Return the namespaces of the class"""
    return NamespaceManager[cls]


def build(
        namespace: HttpUrl,
        namespace_prefix: str,
        class_name: str,
        properties: List[Union[Property, Dict]],
        baseclass=Thing) -> Type[Thing]:
    """Build a ThingModel class

    Parameters
    ----------
    namespace: str
        The namespace of the class
    namespace_prefix: str
        The namespace prefix of the class
    class_name: str
        The name of the class
    properties: Dict[str, Union[str, int, float, bool, datetime, BlankNodeType, None]]
        The properties of the class
    baseclass: Type[Thing]
        The base class to inherit from, default is Thing


    Returns
    -------
    Thing
        A Thing
    """
    _properties = []
    for prop in properties:
        if isinstance(prop, dict):
            _properties.append(Property(**prop))
        else:
            _properties.append(prop)

    annotations = {prop.name: prop.property_type for prop in _properties}
    default_values = {prop.name: prop.default for prop in _properties}

    new_cls = type(
        class_name,
        (baseclass,),
        {
            "__annotations__": annotations,  # Define field type
            **default_values,
        }
    )
    from ontolutils.classes.decorator import _decorate_urirefs, _add_namesapces
    _urirefs = {class_name: f"{namespace_prefix}:{class_name}"}
    _namespaces = {namespace_prefix: namespace}
    for prop in _properties:
        _ns = prop.namespace
        _nsp = prop.namespace_prefix
        if _ns is None:
            _ns = namespace
            _nsp = namespace_prefix
        _urirefs[prop.name] = f"{_nsp}:{prop.name}"
        if _nsp not in _namespaces:
            _namespaces[_nsp] = _ns

    _decorate_urirefs(new_cls, **_urirefs)
    _add_namesapces(new_cls, _namespaces)
    return new_cls


def is_semantically_equal(thing1, thing2) -> bool:
    # Prüfe, ob beide Instanzen von Thing sind
    if isinstance(thing1, Thing) and isinstance(thing2, Thing):
        return thing1.uri == thing2.uri
    return thing1 == thing2



import pathlib
import shutil
import unittest
from hashlib import sha256

import rdflib
from pydantic import EmailStr

import ontolutils
from ontolutils import Thing
from ontolutils import namespaces, urirefs
from ontolutils import set_logging_level
from ontolutils.cache import get_cache_dir
from ontolutils.classes import utils

set_logging_level('WARNING')


class TestUtils(unittest.TestCase):

    def test_UNManager(self):
        unm = utils.UNManager()
        self.assertEqual(unm.__repr__(), 'UNManager()')

        @namespaces(foaf="http://xmlns.com/foaf/0.1/")
        @urirefs(Agent='foaf:Agent',
                 mbox='foaf:mbox')
        class Agent(Thing):
            """Pydantic Model for http://xmlns.com/foaf/0.1/Agent
            Parameters
            ----------
            mbox: EmailStr = None
                Email address (foaf:mbox)
            """
            mbox: EmailStr = None

        self.assertEqual({}, unm.get(Agent, {}))
        self.assertEqual(None, unm.get(Agent, None))
        unm[Agent]['foaf'] = 'http://xmlns.com/foaf/0.1/'
        self.assertEqual(unm.__repr__(), f'UNManager({Agent.__name__})')

    def test_split_uriref(self):
        ns, name = utils.split_URIRef('foaf:Agent')
        self.assertEqual(ns, 'foaf')
        self.assertEqual(name, 'Agent')

        ns, name = utils.split_URIRef('http://xmlns.com/foaf/0.1/Agent')
        self.assertEqual(ns, 'http://xmlns.com/foaf/0.1/')
        self.assertEqual(name, 'Agent')

        ns, name = utils.split_URIRef('Agent')
        self.assertEqual(ns, None)
        self.assertEqual(name, 'Agent')

    def test_download(self):
        text_filename = utils.download_file('https://www.iana.org/assignments/media-types/text.csv')

        self.assertIsInstance(text_filename, pathlib.Path)
        self.assertTrue(text_filename.exists())
        self.assertTrue(text_filename.parent.parent == get_cache_dir())

        # compute hash:
        with open(text_filename, 'rb') as f:
            hash = sha256(f.read()).hexdigest()
        text_filename = utils.download_file('https://www.iana.org/assignments/media-types/text.csv',
                                            known_hash=hash,
                                            dest_filename='not/existing/dir')
        self.assertTrue(text_filename.parent, 'dir')

        text_filename = utils.download_file('https://www.iana.org/assignments/media-types/text.csv',
                                            known_hash=hash,
                                            dest_filename='not/existing/dir',
                                            overwrite_existing=False)
        self.assertTrue(text_filename.exists())

        text_filename = utils.download_file('https://www.iana.org/assignments/media-types/text.csv',
                                            known_hash=hash,
                                            dest_filename='not/existing/dir',
                                            overwrite_existing=True)
        self.assertTrue(text_filename.exists())

        text_filename.unlink(missing_ok=True)

        if pathlib.Path('not/existing/dir').exists():
            shutil.rmtree('not/existing/dir')

    def test_parse_qudt_units(self):
        self.assertEqual(ontolutils.parse_unit('m/s'), rdflib.URIRef('http://qudt.org/vocab/unit/M-PER-SEC'))


import unittest

from ontolutils.ex.pimsii import Property, Variable


class TestPIMSII(unittest.TestCase):

    def testVariable(self):
        variable = Variable(
            id="_:b1",
            label="my variable"
        )
        self.assertEqual(variable.serialize("ttl"), """@prefix pims: <http://www.molmod.info/semantics/pims-ii.ttl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .

[] a pims:Variable ;
    rdfs:label "my variable" .

""")

    def testProperty(self):
        prop = Property(
            id="_:b1",
            label="my property",
            hasValue=5.4
        )
        print(prop.serialize("ttl"))
        self.assertEqual(prop.serialize("ttl"), """@prefix m4i: <http://w3id.org/nfdi4ing/metadata4ing#> .
@prefix pims: <http://www.molmod.info/semantics/pims-ii.ttl#> .
@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .
@prefix xsd: <http://www.w3.org/2001/XMLSchema#> .

[] a pims:Property ;
    rdfs:label "my property" ;
    m4i:hasValue 5.4e+00 .

""")



from typing import Dict

CONFIG = {
    'blank_id_generator': None,
    'show_lang_in_str': True
}
_VALIDATORS = {
    'blank_id_generator': lambda x: callable(x) or x is None,
    'show_lang_in_str': lambda x: isinstance(x, bool) or x is None,
}


class set_config:
    """Set the configuration parameters."""

    def __init__(self, **kwargs):
        self.old = {}
        for k, v in kwargs.items():
            if k in _VALIDATORS and not _VALIDATORS[k](v):
                raise ValueError(f'Config parameter "{k}" has invalid value: "{v}"')
            if k not in CONFIG:
                raise KeyError(f'Not a configuration key: "{k}"')
            self.old[k] = CONFIG[k]
        self._update(kwargs)

    def __enter__(self):
        return

    def __exit__(self, *args, **kwargs):
        self._update(self.old)

    def _update(self, options_dict: Dict):
        CONFIG.update(options_dict)


def get_config(key=None):
    """Return the configuration parameters."""
    if key is None:
        return CONFIG
    return CONFIG[key]



"""Classes for working with ontology data"""
from .decorator import namespaces, urirefs
from .query_util import query, dquery
from .thing import Thing, get_urirefs, get_namespaces, build, Property, LangString
from .urivalue import URIValue
from .utils import as_id

__all__ = [
    'namespaces', 'urirefs',
    'Thing', 'LangString',
    'query', 'dquery',
    'get_urirefs', 'get_namespaces',
    'as_id', 'build',
    'Property',
    'URIValue'
]



# https://www.iana.org/assignments/media-types/text.csv
from typing import Dict

from ontolutils.cache import get_cache_dir
from ontolutils.classes.utils import download_file

iana_cache = get_cache_dir() / 'iana'
iana_cache.mkdir(exist_ok=True)


def read_csv_file(filename, prefix: str) -> Dict[str, str]:
    with open(filename) as f:
        lines = f.read().split('\n')
    # header = lines[0].split(',')
    names = []
    url = []
    for line in lines[1:]:
        if line:
            n, t, r = line.split(',', 2)
            names.append(n)
            url.append(prefix + t)
    return {n: u for n, u in zip(names, url)}


def get_media_type(category: str) -> dict:
    """Return a dictionary of media types for the given category."""
    csv_fname = iana_cache / f'{category}.csv'
    if not csv_fname.exists():
        download_link = f'https://www.iana.org/assignments/media-types/{category}.csv'
        csv_fname = download_file(download_link,
                                  dest_filename=iana_cache / f'{category}.csv')
    return read_csv_file(csv_fname, 'https://www.iana.org/assignments/media-types/')


class IANACLS:

    def __init__(self):
        self._application = None
        self._audio = None
        self._font = None
        self._image = None
        self._message = None
        self._model = None
        self._multipart = None
        self._text = None
        self._video = None

    @classmethod
    def get(cls, media_type):
        """Get media type by full type string, e.g., 'text/html'."""
        main_type, mtype = media_type.split('/',1)
        if mtype.startswith('x-'):
            return "https://www.iana.org/assignments/media-types/" + media_type
        instance = cls()
        media_types = getattr(instance, main_type, {})
        return media_types.get(mtype, None)

    @property
    def application(self):
        """Return a dictionary of application media types."""
        if self._application is None:
            self._application = get_media_type('application')
        return self._application

    @property
    def audio(self):
        """Return a dictionary of audio media types."""
        if self._audio is None:
            self._audio = get_media_type('audio')
        return self._audio

    @property
    def font(self):
        """Return a dictionary of font media types."""
        if self._font is None:
            self._font = get_media_type('font')
        return self._font

    @property
    def image(self):
        """Return a dictionary of image media types."""
        if self._image is None:
            self._image = get_media_type('image')
        return self._image

    @property
    def message(self):
        """Return a dictionary of message media types."""
        if self._message is None:
            self._message = get_media_type('message')
        return self._message

    @property
    def model(self):
        """Return a dictionary of model media types."""
        if self._model is None:
            self._model = get_media_type('model')
        return self._model

    @property
    def multipart(self):
        """Return a dictionary of multipart media types."""
        if self._multipart is None:
            self._multipart = get_media_type('multipart')
        return self._multipart

    @property
    def text(self):
        """Return a dictionary of text media types."""
        if self._text is None:
            self._text = get_media_type('text')
        return self._text

    @property
    def video(self):
        """Return a dictionary of video media types."""
        if self._video is None:
            self._video = get_media_type('video')
        return self._video


IANA = IANACLS()



import pathlib
import unittest

__this_dir__ = pathlib.Path(__file__).parent

from ontolutils.ex.spdx import Checksum
from ontolutils.namespacelib.spdx import SPDX


class TestSpdx(unittest.TestCase):

    def test_checksum(self):
        checksum = Checksum(
            algorithm=SPDX.checksumAlgorithm_sha256,
            value='d2d2d2d2d2d2d2d2d2d2'
        )
        self.assertEqual(checksum.algorithm, str(SPDX.checksumAlgorithm_sha256))
        self.assertEqual(checksum.value, 'd2d2d2d2d2d2d2d2d2d2')

        checksum = Checksum(
            algorithm="md5",
            value='a1a1a1a1a1a1a1a1a1a1'
        )
        self.assertEqual(checksum.algorithm, str(SPDX.checksumAlgorithm_md5))
        self.assertEqual(checksum.value, 'a1a1a1a1a1a1a1a1a1a1')
